from __future__ import annotations

import copy
import itertools
import logging
import os
import re
from collections import defaultdict
from pathlib import Path
from typing import (
    Any,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

import pcbnew

from .board_modifier import (
    KICAD_VERSION,
    BoardModifier,
    calculate_distance_matrix,
    get_closest_pads_on_same_net,
    get_common_nets,
    get_distance,
    get_footprint,
    get_optional_footprint,
    get_orientation,
    get_pads_by_net,
    get_position,
    get_side,
    position_in_rotated_coordinates,
    prim_mst,
    reset_rotation,
    rotate,
    set_position,
    set_rotation,
    set_side,
)
from .element_position import ElementInfo, ElementPosition, PositionOption
from .kle_serial import Key, Keyboard, MatrixAnnotatedKeyboard, get_keyboard_from_file
from .plugin_error import PluginError

logger = logging.getLogger(__name__)


class KeyMatrix:
    def __init__(self, board: pcbnew.BOARD, key_format: str, diode_format: str) -> None:
        self.key_format = key_format
        self.diode_format = diode_format

        self._row_format = ""
        self._column_format = ""

        self._switches: Dict[str, pcbnew.FOOTPRINT] = {}
        self._switches_references_by_net: Dict[FrozenSet[str], List[str]] = defaultdict(
            list
        )
        self._diodes_by_switch: Dict[str, List[pcbnew.FOOTPRINT]] = defaultdict(list)

        diodes: List[pcbnew.FOOTPRINT] = []

        switches_nets: Dict[str, Set[str]] = defaultdict(set)
        all_switches_nets: Set[str] = set()
        diodes_nets_by_reference: Dict[str, Set[str]] = defaultdict(set)
        diodes_unique_nets: Dict[str, Set[str]] = {}

        self.key_pattern = re.compile(key_format.format("(.*)"))
        self.diode_pattern = re.compile(diode_format.format("(.*)"))

        def _get_nets(f: pcbnew.FOOTPRINT) -> List[str]:
            return [p.GetNetname() for p in f.Pads() if p.GetNetname() != ""]

        for f in board.GetFootprints():
            reference = f.GetReference()
            if re.match(self.key_pattern, reference):
                self._switches[reference] = f
                nets = _get_nets(f)
                switches_nets[reference].update(nets)
                all_switches_nets.update(nets)
            elif re.match(self.diode_pattern, reference):
                diodes.append(f)
                diodes_nets_by_reference[reference].update(_get_nets(f))

        # reduce diode_nets to contain only diode-unique nets
        # (i.e. not common with any switch)
        for k, v in diodes_nets_by_reference.items():
            diodes_unique_nets[k] = v.difference(all_switches_nets)

        # each switch can have 0 or more diodes
        for f in diodes:
            reference = f.GetReference()
            for switch_reference, key in self._switches.items():
                diodes_nets = diodes_nets_by_reference[reference]
                switch_nets = [p.GetNetname() for p in key.Pads()]
                if common_nets := list(diodes_nets.intersection(switch_nets)):
                    self._diodes_by_switch[switch_reference].append(f)
                    # remove common switch-diode net and add diode-unique net instead,
                    # this way we should get key-matrix nets:
                    switches_nets[switch_reference].discard(*common_nets)
                    switches_nets[switch_reference].update(
                        diodes_unique_nets[reference]
                    )

        for k, v in switches_nets.items():
            if len(list(v)) == 2:
                self._switches_references_by_net[frozenset(v)].append(k)
            else:
                logger.warning(
                    "Unexpected switch net position detected, "
                    "each switch should have two unique nets unambiguously defining "
                    "position in key matrix, switch-by-matrix association can't be used"
                )
                self._switches_references_by_net = {}
                break
        logger.debug(f"Switches by nets: {self._switches_references_by_net}")
        diodes_by_switch = {
            k: [f.GetReference() for f in v] for k, v in self._diodes_by_switch.items()
        }
        logger.debug(f"Diodes by switch: {diodes_by_switch}")

    def first_switch_reference(self) -> str:
        return min(self._switches.keys())

    def number_of_switches(self) -> int:
        return len(self._switches)

    def switch_by_reference(self, reference: str) -> pcbnew.FOOTPRINT:
        return self._switches[reference]

    def switch_by_format_value(self, value: Any) -> pcbnew.FOOTPRINT:
        return self._switches[self.key_format.format(value)]

    def switches_by_reference(self) -> Iterable[Tuple[str, pcbnew.FOOTPRINT]]:
        return self._switches.items()

    def switches_by_reference_ordered(self) -> Iterable[Tuple[str, pcbnew.FOOTPRINT]]:
        def _tryint(s: str) -> Union[str, int]:
            try:
                return int(s)
            except ValueError:
                return s

        def _alphanum_keys(item: Tuple[str, pcbnew.FOOTPRINT]) -> List[Union[str, int]]:
            return [_tryint(c) for c in re.split("([0-9]+)", item[0])]

        return sorted(self._switches.items(), key=_alphanum_keys)

    def is_matrix_ok(self) -> bool:
        return len(self._switches_references_by_net) != 0

    def __guess_format(self, guesses: List[str]) -> str:
        for guess in guesses:
            pattern = re.compile(guess.format("(\\d)+"))
            for net in list(self.matrix_nets()):
                if re.match(pattern, net):
                    return guess
        # out of luck, getting switches by row,column annotation won't work
        return ""

    @property
    def row_format(self) -> str:
        if not self._row_format:
            self._row_format = self.__guess_format(["ROW{}", "R{}"])
        return self._row_format

    @property
    def column_format(self) -> str:
        if not self._column_format:
            self._column_format = self.__guess_format(["COLUMN{}", "COL{}", "C{}"])
        return self._column_format

    def switches_references_by_coordinates(self, row: int, column: int) -> List[str]:
        return self.switches_references_by_netnames(
            self.row_format.format(row), self.column_format.format(column)
        )

    def switches_references_by_netnames(
        self, row_net: str, column_net: str
    ) -> List[str]:
        nets = (row_net, column_net)
        return self._switches_references_by_net[frozenset(nets)]

    def switches_references_by_netname(self, netname: str) -> List[str]:
        result = []
        for nets, switches in self._switches_references_by_net.items():
            if netname in nets:
                result.extend(switches)
        return result

    def diodes_by_switch_reference(self, reference: str) -> List[pcbnew.FOOTPRINT]:
        return self._diodes_by_switch[reference]

    def any_switch_with_multiple_diodes(self) -> bool:
        return any(len(d) > 1 for d in self._diodes_by_switch.values())

    def matrix_nets(self) -> Set[str]:
        return set().union(*self._switches_references_by_net.keys())

    def matrix_rows(self) -> Set[str]:
        pattern = re.compile(self.row_format.format("(\\d)+"))
        return set(filter(lambda net: re.match(pattern, net), self.matrix_nets()))

    def matrix_columns(self) -> Set[str]:
        pattern = re.compile(self.column_format.format("(\\d)+"))
        return set(filter(lambda net: re.match(pattern, net), self.matrix_nets()))


class KeyboardSwitchIterator:
    EXPLICIT_ANNOTATION_LABEL = 10

    def __init__(
        self,
        keyboard: Keyboard,
        key_matrix: KeyMatrix,
    ) -> None:
        self._keyboard = keyboard
        self._key_matrix = key_matrix
        self._explicit_annotations = self.__check_explicit_annotations(keyboard)

    def __check_explicit_annotations(self, keyboard: Keyboard) -> bool:
        number_of_explicit_annotations = sum(
            str(k.get_label(self.EXPLICIT_ANNOTATION_LABEL)).isdigit()
            for k in keyboard.keys
        )
        return number_of_explicit_annotations == len(keyboard.keys)

    def __iter__(self):
        self._keys = iter(self._keyboard.keys)
        self._current_key = 1
        return self

    def __get_footprint(self, key: Key) -> pcbnew.FOOTPRINT:
        if self._explicit_annotations:
            label = key.labels[self.EXPLICIT_ANNOTATION_LABEL]
            try:
                sw = self._key_matrix.switch_by_format_value(label)
            except KeyError as e:
                msg = (
                    "Provided keyboard layout uses explicit annotations "
                    f"but could not find switch {e}, aborting."
                )
                raise PluginError(msg)
        else:
            try:
                sw = self._key_matrix.switch_by_format_value(self._current_key)
            except KeyError as e:
                msg = (
                    f"Could not find switch {e}, aborting.\n"
                    f"Provided keyboard layout requires {len(self._keyboard.keys)} "
                    f"keys, found {self._key_matrix.number_of_switches()} "
                    f"footprints with '{self._key_matrix.key_format}' annotation."
                )
                raise PluginError(msg)
            self._current_key += 1
        return sw

    def __next__(self):
        key = next(self._keys)
        if key:
            if key.decal:
                return self.__next__()
            return key, self.__get_footprint(key)
        else:
            raise StopIteration


class MatrixAnnotatedKeyboardSwitchIterator:
    def __init__(
        self,
        keyboard: MatrixAnnotatedKeyboard,
        key_matrix: KeyMatrix,
    ) -> None:
        self._keyboard = keyboard
        self._key_matrix = key_matrix

    def __iter__(self):
        self._keys = self._keyboard.key_iterator(ignore_alternative=False)
        self._seen: List[Tuple[str, str]] = []
        return self

    def __get_footprint(self, key: Key) -> Optional[pcbnew.FOOTPRINT]:
        matrix_coordinates = MatrixAnnotatedKeyboard.get_matrix_position(key)
        if all(c.isdigit() for c in matrix_coordinates):
            switches = self._key_matrix.switches_references_by_coordinates(
                *map(int, matrix_coordinates)
            )
        else:
            switches = self._key_matrix.switches_references_by_netnames(
                *matrix_coordinates
            )
        switches = sorted(switches)
        logger.debug(f"Got {switches} for {matrix_coordinates} position")
        # assume that alternative keys have same annotation with
        # some sort of suffix so after sorting
        # the option index would get us correct footprint
        try:
            # due to layout collapsing the choices might be missing,
            # instead of using choice value use number of already seen switches
            switch = switches[self._seen.count(matrix_coordinates)]
            fp = self._key_matrix.switch_by_reference(switch)
            self._seen.append(matrix_coordinates)
            return fp
        except Exception:
            logger.warning("Could not find alternative layout footprint")
            return None

    def __next__(self):
        key = next(self._keys)
        if key:
            if key.decal:
                return self.__next__()
            if not (footprint := self.__get_footprint(key)):
                return self.__next__()
            return key, footprint
        else:
            raise StopIteration


def get_key_iterator(
    keyboard: Keyboard,
    key_matrix: KeyMatrix,
) -> Iterator:
    if isinstance(keyboard, MatrixAnnotatedKeyboard):
        if not key_matrix.is_matrix_ok():
            msg = (
                "Detected layout file with via-annotated matrix positions "
                "while not all footprints on PCB can be unambiguously associated "
                "with row/column position.\n"
                "Either net names are unrecognized, netlist is invalid or "
                "or using direct-pin switch connections.\n"
                "Fix netlist problems or use layout file with 'explicit annotation'.\n"
                "For details see https://github.com/adamws/kicad-kbplacer/"
                "blob/master/docs/annotation_guide.md"
            )
            raise PluginError(msg)
        _iter = MatrixAnnotatedKeyboardSwitchIterator(keyboard, key_matrix)
    else:
        _iter = KeyboardSwitchIterator(keyboard, key_matrix)
    return iter(_iter)


class KeyPlacer(BoardModifier):
    def __init__(
        self,
        board: pcbnew.BOARD,
        key_distance: Tuple[float, float] = (19.05, 19.05),
    ) -> None:
        super().__init__(board)

        self.__key_distance_x = cast(int, pcbnew.FromMM(key_distance[0]))
        self.__key_distance_y = cast(int, pcbnew.FromMM(key_distance[1]))

        logger.debug(
            f"Set key 1U distance: {self.__key_distance_x}/{self.__key_distance_y}"
        )

    def apply_switch_connection_template(
        self,
        switch: pcbnew.FOOTPRINT,
        angle: float,
        template_connection: List[pcbnew.PCB_TRACK],
    ) -> None:
        """
        :param switch: Switch footprint to be routed.
        :param angle: Rotation angle (in degrees) of switch footprint
                      (diode rotation is assumed to be the same)
        :param template_connection: List of template elements (tracks and vias) for
                                    routing switch and diode pads. Normalised to
                                    switch position coordinate. Templates
                                    items must not have netcodes assigned.
        """
        logger.info("Using template replication method")
        if angle != 0:
            logger.info(f"Routing at {angle} degree angle")
        switch_position = get_position(switch)
        rejects = []
        for item in template_connection:
            # item is either PCB_TRACK or PCB_VIA, since via extends track
            # we should be safe with `Cast_to_PCB_TRACK` (not doing any via
            # specific operations here)
            track = pcbnew.Cast_to_PCB_TRACK(item)
            new_track = track.Duplicate()
            if KICAD_VERSION < (7, 0, 0):
                new_track.Move(pcbnew.wxPoint(switch_position.x, switch_position.y))
            else:
                new_track.Move(switch_position)
            if angle != 0:
                rotate(new_track, switch_position, angle)
            result = self.add_track_to_board(new_track)
            # depending on the order of track elements placement, some may not pass
            # collision check (for example when starting from middle segment,
            # if it ends to close to a pad). To avoid rejecting false positives,
            # collect rejects and give them a second chance. Only the 'middle' segments
            # (i.e. not starting in a pad) should be in this list, others should
            # get placed properly (unless indeed colliding with some footprints,
            # for example optional stabilizer holes), so running placement of rejects
            # a second time should succeed.
            if result is None:
                rejects.append(new_track)
        for item in rejects:
            self.add_track_to_board(item)

    def route_switch_with_diode(
        self,
        switch: pcbnew.FOOTPRINT,
        diodes: List[pcbnew.FOOTPRINT],
    ) -> None:
        """Performs routing between switch and diode elements.
        It uses two closest (to each other) pads of the same net.

        :param switch: Switch footprint to be routed.
        :param diodes: Diodes footprints to be routed.
        """
        for diode in diodes:
            logger.info(f"Routing {switch.GetReference()} with {diode.GetReference()}")
            if result := get_closest_pads_on_same_net(switch, diode):
                logger.info("Using internal autorouter method")
                switch_pad, diode_pad = result
                self.route(switch_pad, diode_pad)
            else:
                logger.error("Could not find pads with the same net, routing skipped")

    def get_current_relative_element_position(
        self, element1: pcbnew.FOOTPRINT, element2: pcbnew.FOOTPRINT
    ) -> ElementPosition:
        """Returns position of element2 in relation to element1
        in element1 coordinate system (i.e centered in element1 center
        and rotated by its rotation)
        """
        pos1 = get_position(element1)
        pos2 = get_position(element2)
        rot1 = get_orientation(element1)
        rot2 = get_orientation(element2)
        if rot1:
            pos1 = position_in_rotated_coordinates(pos1, -rot1)
            pos2 = position_in_rotated_coordinates(pos2, -rot1)

        x = cast(float, pcbnew.ToMM(pos2.x - pos1.x))
        y = cast(float, pcbnew.ToMM(pos2.y - pos1.y))

        return ElementPosition(
            x,
            y,
            rot2 - rot1,
            get_side(element2),
        )

    def remove_dangling_tracks(self) -> None:
        logger.info("Removing dangling tracks")
        connectivity = self.get_connectivity()

        any_removed = False

        def _is_dangling(track: pcbnew.PCB_TRACK) -> bool:
            if KICAD_VERSION >= (7, 0, 7):
                return connectivity.TestTrackEndpointDangling(track, False)
            return connectivity.TestTrackEndpointDangling(track)

        for track in self.board.GetTracks():
            if _is_dangling(track):
                logger.info(f"Removing {track.m_Uuid.AsString()}")
                self.board.RemoveNative(track)
                any_removed = True

        if any_removed:
            self.remove_dangling_tracks()

    def save_connection_template(
        self,
        switch: pcbnew.FOOTPRINT,
        diodes: List[pcbnew.FOOTPRINT],
        connections: List[pcbnew.PCB_TRACK],
        destination_path: str,
    ) -> None:
        logger.info(f"Saving template to {destination_path}")
        # can't use `CreateEmptyBoard` when running inside KiCad.
        # We want new board file but without new project file,
        # looks like this is not possible with pcbnew API.
        # So create board with project
        board = pcbnew.NewBoard(destination_path)
        # and delete project file
        os.remove(Path(destination_path).with_suffix(".kicad_pro"))

        switch_copy = pcbnew.Cast_to_FOOTPRINT(switch.Duplicate())
        reset_rotation(switch_copy)
        set_position(switch_copy, pcbnew.VECTOR2I(0, 0))

        origin = get_position(switch)
        diode_copies = []
        for d in diodes:
            diode_copy = pcbnew.Cast_to_FOOTPRINT(d.Duplicate())
            if angle := get_orientation(switch):
                rotate(diode_copy, origin, angle)
            set_position(
                diode_copy,
                get_position(diode_copy) - origin,
            )
            diode_copies.append(diode_copy)

        for p in itertools.chain(
            switch_copy.Pads(),
            itertools.chain.from_iterable((d.Pads() for d in diode_copies)),
        ):
            if p.GetNetCode() != 0:
                logger.info(
                    f"Adding net {p.GetNetname()} with netcode {p.GetNetCode()}"
                )
                # adding nets to new board will get them new autoassigned netcodes
                # (the one set at `NETINFO_ITEM` constructor will be discarded)
                net = pcbnew.NETINFO_ITEM(board, p.GetNetname(), -1)
                board.Add(net)

        nets = board.GetNetsByName()
        # synchronize net codes
        for p in itertools.chain(
            switch_copy.Pads(),
            itertools.chain.from_iterable((d.Pads() for d in diode_copies)),
        ):
            if p.GetNetCode() != 0:
                before = p.GetNetCode()
                p.SetNet(nets[p.GetNetname()])
                logger.info(
                    f"Updating pad '{p.GetParentAsString()}:{p.GetPadName()}' "
                    f"net {p.GetNetname()} netcode: {before} -> {p.GetNetCode()}"
                )

        board.Add(switch_copy)
        for d in diode_copies:
            board.Add(d)

        for item in connections:
            # using `Duplicate` here to not alter net assignments of original tracks
            # (which needs to be empty in order to work as template)
            board.Add(item.Duplicate())
        pcbnew.SaveBoard(destination_path, board, aSkipSettings=True)

    def get_connection_template(
        self, key_format: str, diode_format: str, destination_path: str, route: bool
    ) -> List[pcbnew.PCB_TRACK]:
        """Returns list of tracks (including vias) connecting first element
        with reference `key_format` to itself or any other element
        and optionally save it to new `pcbnew` template file.
        The coordinates of returned elements are normalized to center of `key_format`
        element. If `key_format` element is rotated, resulting coordinates are rotated
        back so the template is always in natural (0) orientation.
        """
        switch = get_footprint(self.board, key_format.format(1))

        logger.info(
            "Looking for connection template between "
            f"{switch.GetReference()} and other elements"
        )
        result = []
        origin = get_position(switch)

        connectivity = self.get_connectivity()
        _tracks: Dict[str, pcbnew.PCB_TRACE_T] = {}

        def _get_connected_tracks(item: pcbnew.BOARD_CONNECTED_ITEM) -> None:
            for t in connectivity.GetConnectedTracks(item):
                uid = t.m_Uuid.AsString()
                if uid not in _tracks:
                    _tracks[uid] = t
                    _get_connected_tracks(t)

        for p in switch.Pads():
            _get_connected_tracks(p)

        for item in _tracks.values():
            item_copy = item.Duplicate()
            item_copy.SetNetCode(0)
            if angle := get_orientation(switch):
                rotate(item_copy, origin, angle)
            if KICAD_VERSION < (7, 0, 0):
                item_copy.Move(pcbnew.wxPoint(-origin.x, -origin.y))
            else:
                item_copy.Move(origin * -1)

            if route:
                self.board.RemoveNative(item)
            result.append(item_copy)

        def _format_item(item: pcbnew.PCB_TRACK) -> str:
            if KICAD_VERSION >= (7, 0, 0):
                name = item.GetFriendlyName()
            else:
                name = "Via" if item.Type() == pcbnew.PCB_VIA_T else "Track"
            start = item.GetStart()
            end = item.GetEnd()
            return f"{name} [{start} {end}]"

        items_str = ", ".join([_format_item(i) for i in result])
        logger.info(f"Got connection template: {items_str}")

        if destination_path:
            pattern = re.compile(diode_format.format("(\\d)+"))
            diodes = [
                f
                for f in self.board.GetFootprints()
                if get_common_nets(f, switch) and re.match(pattern, f.GetReference())
            ]
            self.save_connection_template(switch, diodes, result, destination_path)

        return result

    def load_connection_preset(
        self, key_format: str, diode_format: str, source_path: str
    ) -> List[pcbnew.PCB_TRACK]:
        board = pcbnew.LoadBoard(source_path)
        tracks = board.GetTracks()
        for t in tracks:
            t.SetNetCode(0)
        return tracks

    def _calculate_reference_coordinate(
        self,
        keyboard: Keyboard,
        key_matrix: KeyMatrix,
    ) -> pcbnew.VECTOR2I:
        """Calculates value of offset vector to be applied to key coordinates
        in order to align first key center with defined x/y grid
        """

        def _offset(
            grid: Union[int, float],
            coordinate: Union[int, float],
            size: Union[int, float],
        ) -> int:
            if not grid:
                return 0
            pos = (grid * coordinate) + (grid * size // 2)
            quotient = pos // grid
            # add to avoid placing next to drawing sheet borders (it just looks better):
            margin = 2
            return int(grid * (quotient + margin) - pos)

        offset_x = 0
        offset_y = 0
        key_iterator: Iterator = get_key_iterator(keyboard, key_matrix)
        first_key, _ = next(key_iterator)
        if first_key:
            offset_x = _offset(self.__key_distance_x, first_key.x, first_key.width)
            offset_y = _offset(self.__key_distance_y, first_key.y, first_key.height)
        return pcbnew.VECTOR2I(offset_x, offset_y)

    def place_switches(
        self,
        keyboard: Keyboard,
        key_matrix: KeyMatrix,
        key_position: Optional[ElementPosition],
    ) -> None:
        offset = self._calculate_reference_coordinate(keyboard, key_matrix)
        logger.debug(f"Layout offset: {offset}")
        key_iterator: Iterator = get_key_iterator(keyboard, key_matrix)
        for key, switch_footprint in key_iterator:
            reset_rotation(switch_footprint)
            if key_position:
                set_side(switch_footprint, key_position.side)
                set_rotation(switch_footprint, key_position.orientation)

            position = (
                pcbnew.VECTOR2I(
                    int(self.__key_distance_x * (key.x + key.width / 2)),
                    int(self.__key_distance_y * (key.y + key.height / 2)),
                )
                + offset
            )
            set_position(switch_footprint, position)

            angle = key.rotation_angle
            if angle != 0:
                rotation_reference = (
                    pcbnew.VECTOR2I(
                        int(self.__key_distance_x * key.rotation_x),
                        int(self.__key_distance_y * key.rotation_y),
                    )
                    + offset
                )
                rotate(switch_footprint, rotation_reference, angle)

    def place_element(
        self,
        footprint: pcbnew.FOOTPRINT,
        element_position: ElementPosition,
        reference_position: pcbnew.VECTOR2I,
        reference_orientation: float,
    ) -> None:
        reset_rotation(footprint)
        set_side(footprint, element_position.side)
        set_rotation(footprint, element_position.orientation)

        offset = pcbnew.VECTOR2I_MM(element_position.x, element_position.y)
        if reference_orientation != 0:
            offset = position_in_rotated_coordinates(offset, reference_orientation)

        set_position(footprint, reference_position + offset)
        if reference_orientation != 0:
            current_position = get_position(footprint)
            rotate(footprint, current_position, -1 * reference_orientation)

    def place_diodes(
        self,
        diode_infos: List[ElementInfo],
        key_matrix: KeyMatrix,
    ) -> None:
        if diode_infos and diode_infos[0].position:
            # one diode can be associated with more than one switch,
            # for example for via-annotated alternative keys or when using
            # SOT-23 package or similar, keep track of placed diodes to avoid
            # placing one diode twice.
            placed_diodes: List[pcbnew.FOOTPRINT] = []
            for (
                reference,
                switch_footprint,
            ) in key_matrix.switches_by_reference_ordered():
                diodes = key_matrix.diodes_by_switch_reference(reference)
                switch_position = get_position(switch_footprint)
                switch_orientation = get_orientation(switch_footprint)
                for diode, info in zip(diodes, diode_infos):
                    if info.position and diode not in placed_diodes:
                        self.place_element(
                            diode,
                            info.position,
                            switch_position,
                            switch_orientation,
                        )
                        placed_diodes.append(diode)

    def optimize_diodes_orientation(
        self,
        key_matrix: KeyMatrix,
    ) -> None:
        for (
            reference,
            switch,
        ) in key_matrix.switches_by_reference_ordered():
            diodes = key_matrix.diodes_by_switch_reference(reference)
            logger.debug(f"Optimizing orientation of {switch.GetReference()} diodes")
            for diode in diodes:
                if pads := get_closest_pads_on_same_net(diode, switch):
                    distance1 = get_distance(*pads)
                    position = get_position(diode)
                    rotate(diode, position, 180)
                    distance2 = get_distance(*pads)
                    diff = abs(distance1 - distance2)
                    # if first was shorter or approximately the same (less than 0.01mm)
                    # then go back to initial orientation
                    if distance1 < distance2 or diff < 10000:
                        rotate(diode, position, 180)
                    else:
                        logger.debug(
                            f"Rotated {diode.GetReference()} to minimize distance"
                        )
                else:
                    logger.error(
                        "Could not find pads with the same net, "
                        "diode optimization skipped"
                    )

    def place_switch_elements(
        self,
        elements: List[ElementInfo],
        key_matrix: KeyMatrix,
    ) -> None:
        for reference, switch_footprint in key_matrix.switches_by_reference():
            logger.debug(f"Placing additional elements for {reference}")
            switch_position = get_position(switch_footprint)
            switch_orientation = get_orientation(switch_footprint)
            match = re.match(key_matrix.key_pattern, reference)
            # ommit reference prefix in order to properly match additional element
            # i.e SW1 -> LED1, SW20_1 -> ST20_1
            reference_value = match.group(1) if match else ""
            for element_info in elements:
                footprint = get_optional_footprint(
                    self.board,
                    element_info.annotation_format.format(reference_value),
                )
                if footprint and element_info.position:
                    self.place_element(
                        footprint,
                        element_info.position,
                        switch_position,
                        switch_orientation,
                    )

    def route_switches_with_diodes(
        self,
        key_matrix: KeyMatrix,
        template_connection: List[pcbnew.PCB_TRACK],
    ) -> None:
        if template_connection:
            for _, switch_footprint in key_matrix.switches_by_reference():
                angle = -1 * switch_footprint.GetOrientationDegrees()
                self.apply_switch_connection_template(
                    switch_footprint, angle, template_connection
                )
            # when done, delete all template items
            for item in template_connection:
                self.board.RemoveNative(item)
        else:
            for reference, switch_footprint in key_matrix.switches_by_reference():
                diodes = key_matrix.diodes_by_switch_reference(reference)
                self.route_switch_with_diode(switch_footprint, diodes)

    def route_rows_and_columns(self, key_matrix: KeyMatrix) -> None:
        pads = get_pads_by_net(self.board)

        matrix_net_names = key_matrix.matrix_nets()
        matrix_pads = {net: pads[net] for net in matrix_net_names}
        for net, pads in matrix_pads.items():
            logger.debug(f"Routing {net} pads")
            distances = calculate_distance_matrix(cast(List[pcbnew.BOARD_ITEM], pads))
            result = prim_mst(distances)
            for i, j in result:
                p1 = pads[i]
                p2 = pads[j]
                if p1.GetParentAsString() != p2.GetParentAsString():
                    self.route(p1, p2)

    def load_template(self, template_path: str) -> pcbnew.BOARD:
        if KICAD_VERSION >= (8, 0, 0):
            return pcbnew.PCB_IO_MGR.Load(pcbnew.PCB_IO_MGR.KICAD_SEXP, template_path)  # type: ignore
        return pcbnew.IO_MGR.Load(pcbnew.IO_MGR.KICAD_SEXP, template_path)  # type: ignore

    def _normalize_template_path(self, template_path: str) -> str:
        if not template_path:
            msg = "Template path can't be empty"
            raise PluginError(msg)
        if str(Path(template_path).name) == template_path:
            # if passed filename without directory,
            # assume that this refers to file in the directory of
            # current board. This is mostly for CLI convenience
            template_path = str(Path(self.board.GetFileName()).parent / template_path)
        return template_path

    def _get_relative_position_source(self, element: ElementInfo) -> pcbnew.BOARD:
        if element.position_option == PositionOption.PRESET:
            return self.load_template(
                self._normalize_template_path(element.template_path)
            )
        else:
            return self.board

    def _update_element_position(
        self, key_info: ElementInfo, element: ElementInfo
    ) -> None:
        if element.position_option in [
            PositionOption.RELATIVE,
            PositionOption.PRESET,
        ]:
            source = self._get_relative_position_source(element)
            element1 = get_footprint(source, key_info.annotation_format.format(1))
            element2 = get_footprint(source, element.annotation_format.format(1))
            element.position = self.get_current_relative_element_position(
                element1, element2
            )
            logger.info(f"Element info updated: {element}")

    def _prepare_diode_infos(
        self, key_matrix: KeyMatrix, diode_info: ElementInfo
    ) -> List[ElementInfo]:
        infos = []
        if diode_info.position_option in [
            PositionOption.RELATIVE,
            PositionOption.PRESET,
        ]:
            source = self._get_relative_position_source(diode_info)
            template_matrix = KeyMatrix(
                source, key_matrix.key_format, diode_info.annotation_format
            )
            number_of_template_switches = template_matrix.number_of_switches()
            if (
                diode_info.position_option == PositionOption.PRESET
                and number_of_template_switches != 1
            ):
                msg = (
                    f"Template file '{diode_info.template_path}' "
                    "must have exactly one switch. "
                    f"Found {number_of_template_switches} switches using "
                    f"'{key_matrix.key_format}' annotation format."
                )
                raise PluginError(msg)
            first_switch = template_matrix.first_switch_reference()
            switch = template_matrix.switch_by_reference(first_switch)
            diodes = template_matrix.diodes_by_switch_reference(switch.GetReference())
            for diode in diodes:
                temp_info = copy.copy(diode_info)
                temp_info.position = self.get_current_relative_element_position(
                    switch, diode
                )
                logger.info(f"Element info updated: {temp_info}")
                infos.append(temp_info)
        else:
            infos.append(diode_info)
        return infos

    def _get_template_connection(
        self, key_format: str, diode_info: ElementInfo, route: bool
    ) -> List[pcbnew.PCB_TRACK]:
        if diode_info.position_option in [
            PositionOption.RELATIVE,
            PositionOption.UNCHANGED,
        ]:
            return self.get_connection_template(
                key_format,
                diode_info.annotation_format,
                diode_info.template_path,
                route,
            )
        elif diode_info.position_option == PositionOption.PRESET:
            logger.info(
                f"Loading diode connection preset from {diode_info.template_path}"
            )
            return self.load_connection_preset(
                key_format,
                diode_info.annotation_format,
                self._normalize_template_path(diode_info.template_path),
            )
        else:
            return []

    def run(
        self,
        layout_path: str,
        key_info: ElementInfo,
        diode_info: ElementInfo,
        route_switches_with_diodes: bool = False,
        route_rows_and_columns: bool = False,
        additional_elements: List[ElementInfo] = [],
        optimize_diodes_orientation: bool = False,
    ) -> None:
        # stage 1 - prepare
        key_matrix = KeyMatrix(
            self.board, key_info.annotation_format, diode_info.annotation_format
        )
        if (
            key_matrix.any_switch_with_multiple_diodes()
            and diode_info.position_option
            in [
                PositionOption.DEFAULT,
                PositionOption.CUSTOM,
            ]
        ):
            msg = (
                f"The '{diode_info.position_option}' position not supported for "
                f"multiple diodes per switch layouts, use '{PositionOption.RELATIVE}' "
                f"or '{PositionOption.PRESET}' position option"
            )
            raise PluginError(msg)

        number_of_switches = key_matrix.number_of_switches()
        if number_of_switches == 0:
            msg = (
                f"No switch footprints found using '{key_info.annotation_format}' "
                "annotation format. Make sure that switches are added to opened "
                "PCB file and theirs annotations match configured value."
            )
            raise PluginError(msg)

        # it is important to get template connection
        # and relative positions before moving any elements
        template_connection = self._get_template_connection(
            key_info.annotation_format, diode_info, route_switches_with_diodes
        )

        diode_infos = self._prepare_diode_infos(key_matrix, diode_info)
        for element_info in additional_elements:
            self._update_element_position(key_info, element_info)

        # stage 2 - place elements
        if layout_path:
            keyboard = get_keyboard_from_file(layout_path)
            if not isinstance(keyboard, MatrixAnnotatedKeyboard):
                # if not MatrixAnnotatedKeyboard already,
                # check if it is possible to convert
                try:
                    keyboard = MatrixAnnotatedKeyboard(keyboard.meta, keyboard.keys)
                    logger.info(
                        "Detected layout convertible to matrix annotated keyboard"
                    )
                except Exception:
                    pass
            if isinstance(keyboard, MatrixAnnotatedKeyboard):
                # can be called only once:
                keyboard.collapse()
            self.place_switches(keyboard, key_matrix, key_info.position)

        logger.info(f"Diode info: {diode_infos}")
        if diode_info.position_option != PositionOption.UNCHANGED:
            self.place_diodes(diode_infos, key_matrix)
            if optimize_diodes_orientation:
                self.optimize_diodes_orientation(key_matrix)

        if additional_elements:
            self.place_switch_elements(additional_elements, key_matrix)

        # stage 3 - route elements
        if route_switches_with_diodes:
            self.route_switches_with_diodes(key_matrix, template_connection)

        if route_rows_and_columns:
            self.route_rows_and_columns(key_matrix)

        if route_switches_with_diodes or route_rows_and_columns:
            self.remove_dangling_tracks()



import os
import traceback

import wx

from .plugin_error import PluginError


class ErrorDialog(wx.Dialog):
    def __init__(self, parent, e: Exception) -> None:
        super(ErrorDialog, self).__init__(
            parent, -1, "kbplacer error", style=wx.DEFAULT_DIALOG_STYLE | wx.STAY_ON_TOP
        )
        self.parent = parent
        box = wx.BoxSizer(wx.VERTICAL)

        if type(e) == PluginError:
            message = e.message
            add_traceback = False
        else:
            message = getattr(e, "message", f"{e.__class__.__name__}: {e}")
            add_traceback = True

        message_text = wx.StaticText(self, label=message)

        error_icon = wx.ArtProvider.GetBitmap(
            wx.ART_ERROR, wx.ART_MESSAGE_BOX, wx.Size(32, 32)
        )
        icon = wx.StaticBitmap(self, bitmap=error_icon)

        icon_and_message_sizer = wx.BoxSizer(wx.HORIZONTAL)
        icon_and_message_sizer.Add(icon, 0, wx.ALL, 10)
        icon_and_message_sizer.Add(
            message_text, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 10
        )
        box.Add(icon_and_message_sizer, 0, wx.EXPAND | wx.ALL, 5)

        if add_traceback:
            traceback_str = traceback.format_exc()
            details_text = wx.TextCtrl(
                self,
                value=traceback_str,
                style=wx.TE_MULTILINE | wx.TE_READONLY | wx.HSCROLL,
            )
            self.adjust_size_to_text(details_text, traceback_str)
            box.Add(details_text, 0, wx.EXPAND | wx.ALL, 5)

        buttons = self.CreateButtonSizer(wx.OK)
        box.Add(buttons, 0, wx.EXPAND | wx.ALL, 5)

        self.SetSizerAndFit(box)

    def adjust_size_to_text(self, text_ctrl: wx.TextCtrl, text: str) -> None:
        dc = wx.ClientDC(self)
        max_line_width = 0

        for line in text.split("\n"):
            width, _ = dc.GetTextExtent(line)
            max_line_width = max(max_line_width, width)

        width_for_text = max_line_width + 50

        if self.parent is None:
            max_width = wx.GetDisplaySize().GetWidth() - 100
        else:
            max_width = self.parent.GetSize().GetWidth() - 100

        final_width = min(width_for_text, max_width)

        text_ctrl.SetMinSize(wx.Size(final_width, -1))
        self.SetMinSize(wx.Size(final_width + 20, -1))


if __name__ == "__main__":
    import threading

    try:
        _ = 1 / 0
    except Exception as e:
        app = wx.App()
        dlg = ErrorDialog(None, e)

        if "PYTEST_CURRENT_TEST" in os.environ:
            # use stdin for gracefully closing GUI when running
            # from pytest. This is required when measuring
            # coverage and process kill would cause measurement to be lost
            def listen_for_exit():
                input("Press any key to exit: ")
                dlg.Close()
                wx.Exit()

            input_thread = threading.Thread(target=listen_for_exit)
            input_thread.start()

            dlg.Show()
            app.MainLoop()
        else:
            dlg.ShowModal()

        print("exit ok")



from __future__ import annotations

import logging
import re
from typing import List, cast

import pcbnew

from .board_modifier import KICAD_VERSION

logger = logging.getLogger(__name__)


def convex_hull(points):
    """Computes the convex hull of a set of 2D points.

    Input: an iterable sequence of (x, y) pairs representing the points.
    Output: a list of vertices of the convex hull in counter-clockwise order,
      starting from the vertex with the lexicographically smallest coordinates.
    Implements Andrew's monotone chain algorithm. O(n log n) complexity.

    Source: https://algorithmist.com/wiki/Monotone_chain_convex_hull.py
    """

    # Sort the points lexicographically (tuples are compared lexicographically).
    # Remove duplicates to detect the case we have just one unique point.
    points = sorted(set(points))

    # Boring case: no points or a single point, possibly repeated multiple times.
    if len(points) <= 1:
        return points

    # 2D cross product of OA and OB vectors, i.e. z-component of their 3D cross product.
    # Returns a positive value, if OAB makes a counter-clockwise turn,
    # negative for clockwise turn, and zero if the points are collinear.
    def cross(o, a, b):
        return (a[0] - o[0]) * (b[1] - o[1]) - (a[1] - o[1]) * (b[0] - o[0])

    # Build lower hull
    lower = []
    for p in points:
        while len(lower) >= 2 and cross(lower[-2], lower[-1], p) <= 0:
            lower.pop()
        lower.append(p)

    # Build upper hull
    upper = []
    for p in reversed(points):
        while len(upper) >= 2 and cross(upper[-2], upper[-1], p) <= 0:
            upper.pop()
        upper.append(p)

    # Concatenation of the lower and upper hulls gives the convex hull.
    # Last point of each list is omitted because it is repeated
    # at the beginning of the other list.
    return lower[:-1] + upper[:-1]


def build_board_outline_around_footprints(
    board: pcbnew.BOARD, outline_delta: float, footprints: List[pcbnew.FOOTPRINT]
) -> None:
    if not footprints:
        msg = "Footprints for generating board edge not set"
        raise Exception(msg)

    hulls = [f.GetBoundingHull() for f in footprints]
    points = []
    for hull in hulls:
        for i in range(0, hull.OutlineCount()):
            for p in hull.Outline(i).CPoints():
                points.append((p.x, p.y))
    result = convex_hull(points)
    shape_line = pcbnew.SHAPE_LINE_CHAIN()
    for r in result:
        shape_line.Append(r[0], r[1])

    outline = pcbnew.SHAPE_POLY_SET()
    outline.AddOutline(shape_line)

    def _inflate_outline(outline, delta_mm):
        if KICAD_VERSION < (8, 0, 0):
            outline.Inflate(delta_mm, 10, pcbnew.SHAPE_POLY_SET.CHAMFER_ALL_CORNERS)
        else:
            outline.Inflate(delta_mm, pcbnew.CORNER_STRATEGY_CHAMFER_ALL_CORNERS, 0)

    def _deflate_outline(outline, delta_mm):
        if KICAD_VERSION < (8, 0, 0):
            outline.Deflate(delta_mm, 10, pcbnew.SHAPE_POLY_SET.CHAMFER_ALL_CORNERS)
        else:
            outline.Deflate(delta_mm, pcbnew.CORNER_STRATEGY_CHAMFER_ALL_CORNERS, 0)

    delta_mm = cast(int, pcbnew.FromMM(abs(outline_delta)))
    if outline_delta > 0:
        _inflate_outline(outline, delta_mm)
    elif outline_delta < 0:
        _deflate_outline(outline, delta_mm)

    points = []
    for i in range(0, outline.OutlineCount()):
        for p in outline.Outline(i).CPoints():
            points.append((p.x, p.y))
    # add first point to the end, easier zip iterate for closed shape:
    points.append(points[0])
    for start, end in zip(points, points[1:]):
        segment = pcbnew.PCB_SHAPE(board)
        segment.SetShape(pcbnew.SHAPE_T_SEGMENT)
        segment.SetLayer(pcbnew.Edge_Cuts)
        if KICAD_VERSION < (7, 0, 0):
            segment.SetStart(pcbnew.wxPoint(*start))
            segment.SetEnd(pcbnew.wxPoint(*end))
        else:
            segment.SetStart(pcbnew.VECTOR2I(*start))
            segment.SetEnd(pcbnew.VECTOR2I(*end))
        segment.SetWidth(pcbnew.FromMM(0.4))
        board.Add(segment)


def build_board_outline(
    board: pcbnew.BOARD, outline_delta: float, footprint_annotation_format: str
) -> None:
    selected_footprints = []
    if KICAD_VERSION >= (7, 0, 0):
        selection: pcbnew.DRAWINGS = pcbnew.GetCurrentSelection()
        if len(selection) != 0:
            # else use only selected footprints
            logger.info("Running board edge generation around selected footprints")
            selected_footprints = [
                f.Cast() for f in selection if isinstance(f.Cast(), pcbnew.FOOTPRINT)
            ]
            if not selected_footprints:
                logger.info("No footprints selected")

    if not selected_footprints:
        # if nothing selected use all switches
        logger.info(
            "Running board edge generation around footprints with annotation "
            f"matching '{footprint_annotation_format}' format"
        )
        footprints = board.GetFootprints()
        pattern = re.compile(footprint_annotation_format.format("(\\d)+"))
        selected_footprints = [
            f for f in footprints if re.match(pattern, f.GetReference())
        ]

    build_board_outline_around_footprints(board, outline_delta, selected_footprints)



import inspect
import logging
from logging import NullHandler

try:
    from ._version import __version__
except ImportError:
    __version__ = "not-found"

__license__ = "GPL-3.0-only"
__version__ = __version__

logging.getLogger(__name__).addHandler(NullHandler())
del NullHandler


def __is_in_call_stack(function_name: str, module_name: str) -> bool:
    current_stack = inspect.stack()

    for frame_info in current_stack:
        frame = frame_info.frame
        if frame.f_globals.get("__name__") == module_name:
            if function_name in frame.f_locals or function_name in frame.f_globals:
                return True

    return False


if __is_in_call_stack("LoadPluginModule", "pcbnew"):
    from .kbplacer_plugin_action import KbplacerPluginAction

    KbplacerPluginAction().register()



from __future__ import annotations

import builtins
import itertools
import logging
import math
import re
from collections import defaultdict
from typing import Dict, List, Optional, Tuple

import pcbnew

from .element_position import Side

logger = logging.getLogger(__name__)
version_match = re.search(r"(\d+)\.(\d+)\.(\d+)", pcbnew.Version())
KICAD_VERSION = tuple(map(int, version_match.groups())) if version_match else ()
if KICAD_VERSION == ():
    logger.warning("Could not determine KiCad version")


def position_in_rotated_coordinates(
    point: pcbnew.VECTOR2I, angle: float
) -> pcbnew.VECTOR2I:
    """
    Map position in xy-Cartesian coordinate system to x'y'-Cartesian which
    has same origin but axes are rotated by angle

    :param point: A point to be mapped
    :param angle: Rotation angle (in degrees) of x'y'-Cartesian coordinates
    :type point: pcbnew.VECTOR2I
    :type angle: float
    :return: Result position in x'y'-Cartesian coordinates
    :rtype: pcbnew.VECTOR2I
    """
    x, y = point.x, point.y
    angle = math.radians(angle)
    xr = (x * math.cos(angle)) + (y * math.sin(angle))
    yr = (-x * math.sin(angle)) + (y * math.cos(angle))
    return pcbnew.VECTOR2I(int(xr), int(yr))


def position_in_cartesian_coordinates(
    point: pcbnew.VECTOR2I, angle: float
) -> pcbnew.VECTOR2I:
    """Performs inverse operation to position_in_rotated_coordinates i.e.
    map position in rotated (by angle) x'y'-Cartesian to xy-Cartesian

    :param point: A point to be mapped
    :param angle: Rotation angle (in degrees) of x'y'-Cartesian coordinates
    :type point: pcbnew.VECTOR2I
    :type angle: float
    :return: Result position in xy-Cartesian coordinates
    :rtype: pcbnew.VECTOR2I
    """
    xr, yr = point.x, point.y
    angle = math.radians(angle)
    x = (xr * math.cos(angle)) - (yr * math.sin(angle))
    y = (xr * math.sin(angle)) + (yr * math.cos(angle))
    return pcbnew.VECTOR2I(int(x), int(y))


def get_footprint(board: pcbnew.BOARD, reference: str) -> pcbnew.FOOTPRINT:
    logger.debug(f"Searching for {reference} footprint in {board.GetFileName()}")
    footprint = board.FindFootprintByReference(reference)
    if footprint is None:
        msg = f"Cannot find footprint {reference}"
        raise RuntimeError(msg)
    return footprint


def get_optional_footprint(
    board: pcbnew.BOARD, reference: str
) -> pcbnew.FOOTPRINT | None:
    logger.debug(
        f"Searching for optional {reference} footprint in {board.GetFileName()}"
    )
    footprint = board.FindFootprintByReference(reference)
    if footprint is None:
        logger.debug("Footprint not found")
    return footprint


def set_position(footprint: pcbnew.FOOTPRINT, position: pcbnew.VECTOR2I) -> None:
    logger.debug(f"Setting {footprint.GetReference()} footprint position: {position}")
    if KICAD_VERSION < (7, 0, 0):
        footprint.SetPosition(pcbnew.wxPoint(position.x, position.y))
    else:
        footprint.SetPosition(position)


def set_position_by_points(footprint: pcbnew.FOOTPRINT, x: int, y: int) -> None:
    set_position(footprint, pcbnew.VECTOR2I(x, y))


def get_position(footprint: pcbnew.FOOTPRINT) -> pcbnew.VECTOR2I:
    position = footprint.GetPosition()
    logger.debug(f"Getting {footprint.GetReference()} footprint position: {position}")
    if KICAD_VERSION < (7, 0, 0):
        return pcbnew.VECTOR2I(position)
    return position


def set_side(footprint: pcbnew.FOOTPRINT, side: Side) -> None:
    if side != get_side(footprint):
        footprint.Flip(footprint.GetPosition(), False)


def get_side(footprint: pcbnew.FOOTPRINT) -> Side:
    if footprint.IsFlipped():
        return Side.BACK
    else:
        return Side.FRONT


def set_rotation(footprint: pcbnew.FOOTPRINT, angle: float) -> None:
    footprint.SetOrientationDegrees(angle)


def reset_rotation(footprint: pcbnew.FOOTPRINT) -> None:
    set_rotation(footprint, 0)


def get_orientation(footprint: pcbnew.FOOTPRINT) -> float:
    return footprint.GetOrientationDegrees()


def rotate(
    item: pcbnew.BOARD_ITEM,
    rotation_reference: pcbnew.VECTOR2I,
    angle: float,
) -> None:
    if KICAD_VERSION < (7, 0, 0):
        item.Rotate(
            pcbnew.wxPoint(rotation_reference.x, rotation_reference.y), angle * -10
        )
    else:
        item.Rotate(
            rotation_reference,
            pcbnew.EDA_ANGLE(angle * -1, pcbnew.DEGREES_T),
        )


def get_distance(i1: pcbnew.BOARD_ITEM, i2: pcbnew.BOARD_ITEM) -> int:
    """Calculate distance between two board items"""
    center1 = i1.GetPosition()
    center2 = i2.GetPosition()
    return int(((center1.x - center2.x) ** 2 + (center1.y - center2.y) ** 2) ** 0.5)


def get_common_layers(p1: pcbnew.PAD, p2: pcbnew.PAD) -> list[int]:
    """Returns list of common layer ids for both of the given pads,
    may be empty if no common layers found
    """
    set1 = [layer for layer in p1.GetLayerSet().CuStack()]
    set2 = [layer for layer in p2.GetLayerSet().CuStack()]
    return list(set(set1).intersection(set2))


def get_common_nets(f1: pcbnew.FOOTPRINT, f2: pcbnew.FOOTPRINT) -> list[int]:
    """Returns list of netcodes which are used by both of the
    given footprints, may be empty if no common nets found
    """
    codes1 = [p.GetNetCode() for p in f1.Pads()]
    codes2 = [p.GetNetCode() for p in f2.Pads()]
    return list(set(codes1).intersection(codes2))


def get_closest(
    pads1: list[pcbnew.PAD], pads2: list[pcbnew.PAD]
) -> Tuple[int, Tuple[pcbnew.PAD, pcbnew.PAD]]:
    all_pairs = itertools.product(pads1, pads2)
    min_distance, closest_pair = min(
        ((get_distance(obj1, obj2), (obj1, obj2)) for obj1, obj2 in all_pairs),
        key=lambda x: x[0],
    )
    return min_distance, closest_pair


def get_closest_pads_on_same_net(
    f1: pcbnew.FOOTPRINT, f2: pcbnew.FOOTPRINT
) -> Tuple[pcbnew.PAD, pcbnew.PAD] | None:
    """Return pair of closest, same net pads of two given footprints
    or None if footprints do not use common net
    """
    pads1 = {}
    for p in f1.Pads():
        if netname := p.GetNetname():
            pads1.setdefault(netname, []).append(p)

    pads2 = {}
    for p in f2.Pads():
        netname = p.GetNetname()
        if netname in pads1:
            pads2.setdefault(netname, []).append(p)

    result = None
    closest = float("inf")
    for key, value in pads2.items():
        distance, result_temp = get_closest(value, pads1[key])
        if distance < closest:
            closest = distance
            result = result_temp

    # We iterated with pads2 so we need to swap result to keep original
    # order in tuple, first pad should be from f1, second from f2
    # Some caller might require proper order
    if result:
        return result[1], result[0]
    return result


def get_clearance(
    i1: pcbnew.BOARD_CONNECTED_ITEM, i2: pcbnew.BOARD_CONNECTED_ITEM
) -> int:
    if KICAD_VERSION < (7, 0, 0):
        # workaround for bug in KiCad 6.x python wrapper:
        import _pcbnew

        def _own_clearance(item: pcbnew.BOARD_CONNECTED_ITEM) -> int:
            if item.Type() == pcbnew.PCB_PAD_T:
                return _pcbnew.PAD_GetOwnClearance(item, item.GetLayer())
            else:
                return _pcbnew.BOARD_CONNECTED_ITEM_GetOwnClearance(
                    item, item.GetLayer()
                )

        return max(_own_clearance(i1), _own_clearance(i2))
    else:
        return max(i1.GetOwnClearance(i1.GetLayer()), i2.GetOwnClearance(i2.GetLayer()))


def get_pads_by_net(board: pcbnew.BOARD) -> Dict[str, List[pcbnew.PAD]]:
    matrix_pads: Dict[str, List[pcbnew.PAD]] = defaultdict(list)

    xy_sorted_pads = pcbnew.PADS_VEC()
    board.GetSortedPadListByXthenYCoord(xy_sorted_pads)

    for pad in xy_sorted_pads:
        net_name = pad.GetNetname()
        matrix_pads[net_name].append(pad)

    return matrix_pads


def calculate_distance_matrix(elements: List[pcbnew.BOARD_ITEM]) -> List[List[int]]:
    distances = []
    for e1 in elements:
        distances1 = []
        for e2 in elements:
            distance = get_distance(e1, e2)
            distances1.append(distance)
        distances.append(distances1)
    return distances


def prim_mst(distances: List[List[int]]) -> List[Tuple[int, int]]:
    """Find minimum spanning tree (MST) using Prim's algorithm"""
    # TODO: KiCad uses Kruskal's algorithm for ratsnests, check which one is better
    # for kbplacer usecase
    n = len(distances)
    in_mst = [False] * n  # Track nodes that are already in MST
    mst_edges = []  # List to store the edges of the MST
    key = [float("inf")] * n  # Used to pick minimum weight edge
    parent = [-1] * n  # Array to store constructed MST

    key[0] = 0  # Start from the first node
    for _ in range(n):
        # Pick the minimum key vertex from the set of vertices not yet in MST
        min_key = float("inf")
        u = -1
        for i in range(n):
            if not in_mst[i] and key[i] < min_key:
                min_key = key[i]
                u = i

        # Include the picked vertex in the MST
        in_mst[u] = True

        # Update key and parent for adjacent vertices of the picked vertex
        for v in range(n):
            if distances[u][v] > 0 and not in_mst[v] and distances[u][v] < key[v]:
                key[v] = distances[u][v]
                parent[v] = u

    # Construct the list of MST edges
    for i in range(1, n):
        if parent[i] != -1:
            mst_edges.append((parent[i], i))  # Add edge (parent, child)

    return mst_edges


def get_netclass(
    board: pcbnew.BOARD, item: pcbnew.BOARD_CONNECTED_ITEM
) -> pcbnew.NETCLASS:
    netclass_name = item.GetNetClassName()
    if KICAD_VERSION < (7, 0, 0):
        netclasses = board.GetNetClasses()
        netclass = netclasses.Find(netclass_name)
        return netclass if netclass else netclasses.Find("Default")
    else:
        # workaround, see https://gitlab.com/kicad/code/kicad/-/issues/18609
        try:
            return board.GetNetClasses()[netclass_name]
        except IndexError:
            # may happen when item has no net assigned yet or netclass is
            # equal "Default" (which is not a part of GetNetClasses collection)
            return board.GetAllNetClasses()["Default"]


class BoardModifier:
    def __init__(self, board: pcbnew.BOARD) -> None:
        self.board = board

    def get_connectivity(self) -> pcbnew.CONNECTIVITY_DATA:
        self.board.BuildConnectivity()
        return self.board.GetConnectivity()

    def _test_collision_track_without_net(self, track: pcbnew.PCB_TRACK) -> bool:
        collide_list = []
        track_shape = track.GetEffectiveShape()
        track_start = track.GetStart()
        track_end = track.GetEnd()
        # connectivity needs to be last,
        # otherwise it will update track net name before we want it to:
        connectivity = self.get_connectivity()
        for f in self.board.GetFootprints():
            reference = f.GetReference()
            hull = f.GetBoundingHull()
            if hit_test_result := hull.Collide(track_shape):
                for p in f.Pads():
                    pad_name = p.GetName()
                    pad_shape = p.GetEffectiveShape()

                    # if track starts or ends in pad then assume that
                    # this collision is expected, with the exception of case
                    # where track already has netlist set and it is different
                    # than pad's netlist or pad has `NPTH` attribute
                    if p.HitTest(track_start) or p.HitTest(track_end):
                        logger.debug(
                            "Track collision ignored, track starts or ends "
                            f"in pad {reference}:{pad_name}"
                        )
                    else:
                        hit_test_result = pad_shape.Collide(
                            track_shape, get_clearance(p, track)
                        )
                        on_same_layer = p.IsOnLayer(track.GetLayer())
                        if hit_test_result and on_same_layer:
                            logger.debug(
                                f"Track collide with pad {reference}:{pad_name}"
                            )
                            collide_list.append(p)
        # track ids to clear at the end:
        tracks_to_clear = []
        for t in self.board.GetTracks():
            # check collision if not itself and on same layer
            if t.m_Uuid != track.m_Uuid and t.IsOnLayer(track.GetLayer()):
                track_uuid = t.m_Uuid.AsString()
                if (
                    track_start == t.GetStart()
                    or track_start == t.GetEnd()
                    or track_end == t.GetStart()
                    or track_end == t.GetEnd()
                ):
                    logger.debug(
                        "Track collision ignored, track starts or ends "
                        f"at the end of {track_uuid} track"
                    )
                    # ignoring one track means that we can ignore
                    # all other connected to it:
                    tracks_to_clear += [
                        x.m_Uuid for x in connectivity.GetConnectedTracks(t)
                    ]
                    # check if connection to this track clears pad collision:
                    connected_pads_ids = [
                        x.m_Uuid for x in connectivity.GetConnectedPads(t)
                    ]
                    for collision in list(collide_list):
                        if collision.m_Uuid in connected_pads_ids:
                            logger.debug(
                                "Pad collision removed due to connection with track "
                                "which leads to that pad"
                            )
                            collide_list.remove(collision)
                elif hit_test_result := t.GetEffectiveShape().Collide(
                    track_shape, get_clearance(t, track)
                ):
                    logger.debug(f"Track collide with another track: {track_uuid}")
                    collide_list.append(t)
        for collision in list(collide_list):
            if collision.m_Uuid in tracks_to_clear:
                collision_uuid = collision.m_Uuid.AsString()
                logger.debug(
                    f"Track collision with {collision_uuid} removed due to "
                    "connection with track which leads to it"
                )
                collide_list.remove(collision)
        return len(collide_list) != 0

    def _test_collision_track_with_net(self, track: pcbnew.PCB_TRACK) -> bool:
        track_netcode = track.GetNetCode()
        track_shape = track.GetEffectiveShape()
        track_layer = track.GetLayer()
        for item in self.board.AllConnectedItems():
            if track_netcode == item.GetNetCode():
                continue

            item_shape = item.GetEffectiveShape()
            if item_shape.Collide(track_shape, get_clearance(item, track)):
                if isinstance(item.Cast(), pcbnew.PCB_TRACK):
                    if item.GetLayer() == track_layer:
                        return True
                elif isinstance(item.Cast(), pcbnew.PAD):
                    pad = pcbnew.Cast_to_PAD(item)
                    if (
                        pad.IsOnLayer(track_layer)
                        or pad.GetAttribute == pcbnew.PAD_ATTRIB_NPTH
                    ):
                        return True
                # pcbnew.ZONE is remaining possibility but we do not handle it
        return False

    def test_track_collision(self, track: pcbnew.PCB_TRACK) -> bool:
        if track.GetNetCode() == 0:
            return self._test_collision_track_without_net(track)
        return self._test_collision_track_with_net(track)

    def add_track_to_board(self, track: pcbnew.PCB_TRACK) -> Optional[pcbnew.VECTOR2I]:
        """Add track to the board if track passes collision check.
        If track has no set netlist, it would get netlist of a pad
        or other track, on which it started or ended.
        Collision with element of the same netlist will be ignored
        unless it is default '0' netlist.
        This exception about '0' netlist is important because it helps
        to detect collisions with holes.

        :param track: A track to be added to board
        :return: End position of added track or None if failed to add.
        """
        layer_name = self.board.GetLayerName(track.GetLayer())
        start = track.GetStart()
        stop = track.GetEnd()
        logger.info(
            f"Adding track segment ({layer_name}): [{start}, {stop}]",
        )
        if not self.test_track_collision(track):
            self.board.Add(track)
            logger.info("Track added")
            if KICAD_VERSION < (7, 0, 0):
                return pcbnew.VECTOR2I(stop.x, stop.y)
            return stop
        else:
            logger.debug("Could not add track segment due to detected collision")
            return None

    def add_track_segment_by_points(
        self,
        start: pcbnew.VECTOR2I,
        end: pcbnew.VECTOR2I,
        *,
        layer: int = pcbnew.B_Cu,
        netcode: int = 0,
        width: int = 200000,
    ) -> Optional[pcbnew.VECTOR2I]:
        track = pcbnew.PCB_TRACK(self.board)
        track.SetWidth(width)
        track.SetLayer(layer)
        if netcode:
            track.SetNetCode(netcode)
        if KICAD_VERSION < (7, 0, 0):
            track.SetStart(pcbnew.wxPoint(start.x, start.y))
            track.SetEnd(pcbnew.wxPoint(end.x, end.y))
        else:
            track.SetStart(start)
            track.SetEnd(end)
        return self.add_track_to_board(track)

    def route(self, pad1: pcbnew.PAD, pad2: pcbnew.PAD) -> bool:
        r"""Simple track router
        If pads are collinear, it will use single track segment.
        If pads are not collinear it will try use two segments track,
        in first try, it will fanout straight segment from `pad1` to
        `pad2` direction and then it will end it with second segment
        at 45degree angle:

        |pad| ____
        | 1 |     \
                   \
                  |pad|
                  | 2 |

        If this does not succeed (it is not possible using 45degree angles
        and two segments only or there is collision detected),
        then it will try to switch track posture:

        |pad|
        | 1 |
          \
           \______|pad|
                  | 2 |

        If this is not possible as well, it will leave pads unconnected
        and will not try another route.

        If both parent footprint are rotated the same, it might
        route track at that angle.
        Footprint is considered rotated, if its orientation is different than
        0, 90, 180 and 270. When comparing rotations use angle to closest
        quarter, i.e. footprints with orientation 10 and 190 (190-180=10) are
        considered to be rotated the same.

        Returns True if routing done, False otherwise
        """
        layers = get_common_layers(pad1, pad2)
        if not layers:
            logger.warning("Could not route pads, no common layers found")
            return False

        layer = layers[0]
        logger.debug(f"Routing at {self.board.GetLayerName(layer)} layer")

        if pad1.GetNetCode() != pad2.GetNetCode():
            logger.warning("Could not route pads of different nets")
            return False

        netcode = pad1.GetNetCode()
        netclass = get_netclass(self.board, pad1)
        track_width = netclass.GetTrackWidth()

        if KICAD_VERSION < (7, 0, 0):
            # on KiCad 6.0.11 `netclass.GetName()` for Default netclass
            # ends with segmentation fault
            netclass_str = f"{netclass}"
        else:
            netclass_str = netclass.GetName()

        logger.debug(f"Netclass: {netclass_str}, track width: {track_width}")

        track_args = {"layer": layer, "netcode": netcode, "width": track_width}

        def _calculate_corners(
            pos1: pcbnew.VECTOR2I, pos2: pcbnew.VECTOR2I
        ) -> Tuple[pcbnew.VECTOR2I, pcbnew.VECTOR2I]:
            x_diff = pos2.x - pos1.x
            y_diff = pos2.y - pos1.y
            x_diff_abs = builtins.abs(x_diff)
            y_diff_abs = builtins.abs(y_diff)
            if x_diff_abs < y_diff_abs:
                up_or_down = -1 if y_diff > 0 else 1
                return (
                    pcbnew.VECTOR2I(pos1.x, pos2.y + (up_or_down * x_diff_abs)),
                    pcbnew.VECTOR2I(pos2.x, pos1.y - (up_or_down * x_diff_abs)),
                )
            else:
                left_or_right = -1 if x_diff > 0 else 1
                return (
                    pcbnew.VECTOR2I(pos2.x + (left_or_right * y_diff_abs), pos1.y),
                    pcbnew.VECTOR2I(pos1.x - (left_or_right * y_diff_abs), pos2.y),
                )

        def _route(
            pos1: pcbnew.VECTOR2I, pos2: pcbnew.VECTOR2I, corner: pcbnew.VECTOR2I
        ) -> bool:
            if end := self.add_track_segment_by_points(pos1, corner, **track_args):
                end = self.add_track_segment_by_points(end, pos2, **track_args)
                return end is not None
            return False

        def _angles_equal(angle1: float, angle2: float) -> bool:
            return abs(angle1 - angle2) <= 0.1

        pos1 = pad1.GetPosition()
        pos2 = pad2.GetPosition()

        parent1 = pad1.GetParent()
        parent2 = pad2.GetParent()

        if KICAD_VERSION >= (8, 0, 0):
            parent1 = pcbnew.Cast_to_FOOTPRINT(parent1)
            parent2 = pcbnew.Cast_to_FOOTPRINT(parent2)

        orientation1 = get_orientation(parent1)
        orientation2 = get_orientation(parent2)

        if _angles_equal(orientation1 % 90, 0) and _angles_equal(orientation2 % 90, 0):
            # this could be merged with next elif but being close to zero is special
            # i.e. we clamp to 0 when near:
            angle = 0
        elif _angles_equal(orientation1 % 90, orientation2 % 90):
            # if rotations are considered the same use the angle
            angle = (-1 * orientation1 + 360) % 360
        else:
            logger.warning(
                "Could not route pads when parent footprints not rotated the same, "
                f"orientations: {orientation1} and {orientation2}"
            )
            return False

        logger.debug(
            f"Routing pad '{pad1.GetParentAsString()}:{pad1.GetPadName()}' at {pos1} "
            f"with pad '{pad2.GetParentAsString()}:{pad2.GetPadName()}' at {pos2} "
            f"using coordinate system rotated by {angle} degree(s)"
        )

        end = None
        # if in line, use one track segment
        if pos1.x == pos2.x or pos1.y == pos2.y:
            end = self.add_track_segment_by_points(pos1, pos2, **track_args)
        else:
            # pads are not in single line, attempt routing with two segment track
            if angle != 0:
                pos1_r = position_in_rotated_coordinates(pos1, angle)
                pos2_r = position_in_rotated_coordinates(pos2, angle)
                corners = _calculate_corners(pos1_r, pos2_r)
                corners = (
                    position_in_cartesian_coordinates(corners[0], angle),
                    position_in_cartesian_coordinates(corners[1], angle),
                )
            else:
                corners = _calculate_corners(pos1, pos2)

            for corner in corners:
                if end := _route(pos1, pos2, corner):
                    break

        return end is not None



from __future__ import annotations

import argparse
import copy
import json
import logging
import pprint
import re
import sys
from collections import defaultdict
from dataclasses import asdict, dataclass, field, fields
from itertools import chain
from typing import Any, Dict, Iterator, List, Optional, Tuple, Type, Union

logger = logging.getLogger(__name__)

DEFAULT_KEY_COLOR = "#cccccc"
DEFAULT_TEXT_COLOR = "#000000"
DEFAULT_TEXT_SIZE = 3
KEY_MAX_LABELS = 12

# Map from serialized label position to normalized position,
# depending on the alignment flags.
# fmt: off
LABEL_MAP: List[List[int]] = [
    # 0  1  2  3  4  5  6  7  8  9 10 11   # align flags
    [ 0, 6, 2, 8, 9,11, 3, 5, 1, 4, 7,10], # 0 = no centering
    [ 1, 7,-1,-1, 9,11, 4,-1,-1,-1,-1,10], # 1 = center x
    [ 3,-1, 5,-1, 9,11,-1,-1, 4,-1,-1,10], # 2 = center y
    [ 4,-1,-1,-1, 9,11,-1,-1,-1,-1,-1,10], # 3 = center x & y
    [ 0, 6, 2, 8,10,-1, 3, 5, 1, 4, 7,-1], # 4 = center front (default)
    [ 1, 7,-1,-1,10,-1, 4,-1,-1,-1,-1,-1], # 5 = center front & x
    [ 3,-1, 5,-1,10,-1,-1,-1, 4,-1,-1,-1], # 6 = center front & y
    [ 4,-1,-1,-1,10,-1,-1,-1,-1,-1,-1,-1], # 7 = center front & x & y
]

REVERSE_LABEL_MAP: List[List[int]] = [
    # 0  1  2  3  4  5  6  7  8  9 10 11   # align flags
    [ 0, 8, 2, 6, 9, 7, 1,10, 3, 4,11, 5], # 0 = no centering
    [-1, 0,-1,-1, 6,-1,-1, 1,-1, 4,11, 5], # 1 = center x
    [-1,-1,-1, 0, 8, 2,-1,-1,-1, 4,11, 5], # 2 = center y
    [-1,-1,-1,-1, 0,-1,-1,-1,-1, 4,11, 5], # 3 = center x & y
    [ 0, 8, 2, 6, 9, 7, 1,10, 3,-1, 4,-1], # 4 = center front (default)
    [-1, 0,-1,-1, 6,-1,-1, 1,-1,-1, 4,-1], # 5 = center front & x
    [-1,-1,-1, 0, 8, 2,-1,-1,-1,-1, 4,-1], # 6 = center front & y
    [-1,-1,-1,-1, 0,-1,-1,-1,-1,-1, 4,-1], # 7 = center front & x & y
]
# fmt: on


@dataclass
class KeyDefault:
    textColor: str = DEFAULT_TEXT_COLOR  # noqa: N815
    textSize: int = DEFAULT_TEXT_SIZE  # noqa: N815


@dataclass
class Key:
    color: str = DEFAULT_KEY_COLOR
    labels: List[Optional[str]] = field(default_factory=list)
    textColor: List[Optional[str]] = field(default_factory=list)  # noqa: N815
    textSize: List[Optional[int]] = field(default_factory=list)  # noqa: N815
    default: KeyDefault = field(default_factory=KeyDefault)
    x: float = 0
    y: float = 0
    width: float = 1
    height: float = 1
    x2: float = 0
    y2: float = 0
    width2: float = 1
    height2: float = 1
    rotation_x: float = 0
    rotation_y: float = 0
    rotation_angle: float = 0
    decal: bool = False
    ghost: bool = False
    stepped: bool = False
    nub: bool = False
    profile: str = ""
    sm: str = ""  # switch mount
    sb: str = ""  # switch brand
    st: str = ""  # switch type

    def __post_init__(self: Key) -> None:
        if isinstance(self.default, dict):
            self.default = KeyDefault(**self.default)
        for key_field in self.__dataclass_fields__:
            value = getattr(self, key_field)
            if isinstance(value, float):
                new_val = round(value, 6)
                setattr(self, key_field, new_val)

    def get_label(self: Key, index: int) -> Optional[str]:
        if len(self.labels) > index:
            return self.labels[index]
        return None

    def set_label(self: Key, index: int, value: str) -> None:
        if index > KEY_MAX_LABELS - 1 or index < 0:
            msg = "Illegal key label index"
            raise RuntimeError(msg)
        labels_to_add = index + 1 - len(self.labels)
        self.labels += labels_to_add * [None]
        self.labels[index] = value


@dataclass
class Background:
    name: str = ""
    style: str = ""


@dataclass
class KeyboardMetadata:
    author: str = ""
    backcolor: str = "#eeeeee"
    background: Optional[Background] = None
    name: str = ""
    notes: str = ""
    radii: str = ""
    switchBrand: str = ""  # noqa: N815
    switchMount: str = ""  # noqa: N815
    switchType: str = ""  # noqa: N815

    def __post_init__(self: KeyboardMetadata) -> None:
        if isinstance(self.background, dict):
            self.background = Background(**self.background)


@dataclass
class Keyboard:
    meta: KeyboardMetadata = field(default_factory=KeyboardMetadata)
    keys: List[Key] = field(default_factory=list)

    @classmethod
    def from_json(cls: Type[Keyboard], data: dict) -> Keyboard:
        if isinstance(data["meta"], dict):
            data["meta"] = KeyboardMetadata(**data["meta"])
        if isinstance(data["keys"], list):
            keys: List[Key] = [Key(**key) for key in data["keys"]]
            data["keys"] = keys
        return cls(**data)

    def to_json(self: Keyboard, indent: Optional[int] = None) -> str:
        return json.dumps(asdict(self), indent=indent)

    def __text_size_changed(self: Keyboard, current: list[Any], new: list[Any]) -> bool:
        current = copy.copy(current)
        new = copy.copy(new)
        for obj in [current, new]:
            if len_difference := KEY_MAX_LABELS - len(obj):
                obj.extend(len_difference * [0])
        return current != new

    def to_kle(self: Keyboard) -> str:
        row = []
        rows = []

        current: Key = copy.deepcopy(Key())
        # some properties are not part of Key type, store them separately:
        current_alignment = 4
        current_f2 = -1
        # rotation origin:
        cluster: dict[str, float] = {"r": 0, "rx": 0, "ry": 0}

        new_row = True
        current.y -= 1  # will be incremented on first row

        for key in self.keys:
            props: dict[str, Any] = {}

            def add_prop(name: str, value: Any, default: Any) -> Any:
                def _round(v: Any) -> Any:
                    return round(v, 6) if isinstance(v, float) else v

                value = _round(value)
                default = _round(default)
                if value != default:
                    props[name] = value
                return value

            if key.labels:
                alignment, labels = find_best_label_alignment(key.labels)
            else:
                alignment, labels = 7, []

            # detect new row
            new_cluster = (
                key.rotation_angle != cluster["r"]
                or key.rotation_x != cluster["rx"]
                or key.rotation_y != cluster["ry"]
            )
            new_row = key.y != current.y
            if row and (new_cluster or new_row):
                # push the old row
                rows.append(row)
                row = []
                new_row = True

            if new_row:
                current.y = round(current.y + 1, 6)
                # 'y' is reset if either 'rx' or 'ry' are changed
                if key.rotation_y != cluster["ry"] or key.rotation_x != cluster["rx"]:
                    current.y = key.rotation_y
                # always reset x to rx (which defaults to zero)
                current.x = key.rotation_x

                cluster["r"] = key.rotation_angle
                cluster["rx"] = key.rotation_x
                cluster["ry"] = key.rotation_y

                new_row = False

            current.rotation_angle = add_prop(
                "r", key.rotation_angle, current.rotation_angle
            )
            current.rotation_x = add_prop("rx", key.rotation_x, current.rotation_x)
            current.rotation_y = add_prop("ry", key.rotation_y, current.rotation_y)

            x_offset = add_prop("x", round(key.x - current.x, 6), 0)
            y_offset = add_prop("y", round(key.y - current.y, 6), 0)
            current.x = round(current.x + key.width + x_offset, 6)
            current.y = round(current.y + y_offset, 6)

            current.color = add_prop("c", key.color, current.color)
            if text_color := reorder_items_kle(key.textColor, alignment):
                if not text_color[0]:
                    text_color[0] = key.default.textColor
                text_color = ["" if not item else item for item in text_color]
                text_color = "\n".join(text_color).rstrip("\n")
                current.textColor = add_prop("t", text_color, current.textColor)
            else:
                current.default.textColor = add_prop(
                    "t", key.default.textColor, current.default.textColor
                )

            current.ghost = add_prop("g", key.ghost, current.ghost)
            current.profile = add_prop("p", key.profile, current.profile)
            current.sm = add_prop("sm", key.sm, current.sm)
            current.sb = add_prop("sb", key.sb, current.sb)
            current.st = add_prop("st", key.st, current.st)

            current_alignment = add_prop("a", alignment, current_alignment)
            current.default.textSize = add_prop(
                "f", key.default.textSize, current.default.textSize
            )
            if "f" in props:
                current.textSize = []

            text_size = reorder_items_kle(key.textSize, alignment)
            text_size = [0 if not isinstance(i, int) else i for i in text_size]
            if self.__text_size_changed(current.textSize, text_size):
                if not text_size:
                    current.default.textSize = add_prop(
                        "f", key.default.textSize, current.default.textSize
                    )
                    current.textSize = []
                else:
                    if optimize := not text_size[0]:
                        optimize = all(x == text_size[1] for x in text_size[2:])
                    if optimize:
                        f2 = text_size[1]
                        current_f2 = add_prop("f2", f2, current_f2)
                        # don't know why this gives type checking error, works fine:
                        current.textSize = [0] + (11 * [f2])  # type: ignore
                    else:
                        current.textSize = add_prop("fa", text_size, [])

            add_prop("w", key.width, 1)
            add_prop("h", key.height, 1)
            add_prop("w2", key.width2, key.width)
            add_prop("h2", key.height2, key.height)
            add_prop("x2", key.x2, 0)
            add_prop("y2", key.y2, 0)
            add_prop("l", key.stepped, False)
            add_prop("n", key.nub, False)
            add_prop("d", key.decal, False)

            if props:
                row.append(props)

            current.labels = labels
            labels = ["" if not item else item for item in labels]
            row.append("\n".join(labels).rstrip("\n"))

        if row:
            rows.append(row)

        result = ""

        default_meta = asdict(KeyboardMetadata())
        meta = copy.deepcopy(asdict(self.meta))
        if meta != default_meta:
            # include only non-default meta fields
            for k in list(meta.keys()):
                if default_meta.get(k, None) == meta[k]:
                    del meta[k]
            result += json.dumps(meta, indent=None) + ",\n"

        for row in rows:
            result += json.dumps(row, indent=None) + ",\n"
        result = result.strip(",\n")
        return result


@dataclass
class MatrixAnnotatedKeyboard(Keyboard):
    MATRIX_COORDINATES_LABEL = 0
    LAYOUT_OPTION_LABEL = 8

    alternative_keys: List[Key] = field(default_factory=list)
    collapsed: bool = field(init=False)

    def __post_init__(self: MatrixAnnotatedKeyboard) -> None:
        positions = []
        for key in list(self.keys):
            if not key.decal:
                # check if required labels defined correctly
                position = MatrixAnnotatedKeyboard.get_matrix_position(key)
                option = MatrixAnnotatedKeyboard.get_layout_option(key)
                if option == 0:
                    positions.append(position)
            if self.__is_alternative(key):
                self.alternative_keys.append(copy.deepcopy(key))
                self.keys.remove(key)
        # check if there are no duplicated matrix position in default key group
        if len(positions) != len(set(positions)):
            msg = "Duplicate matrix position for default layout keys not allowed"
            raise ValueError(msg)

        self.collapsed = False

    def __is_alternative(self, key: Key) -> bool:
        if label := key.get_label(self.LAYOUT_OPTION_LABEL):
            # check if not default layout:
            if label.split(",")[1].strip() != "0":
                # alternative layout key
                return True
        return False

    def key_iterator(self, *, ignore_alternative: bool) -> Iterator[Key]:
        if ignore_alternative:
            return iter(self.keys)
        else:
            return chain(self.keys, self.alternative_keys)

    def _get_layout_option_or_none(self, key: Key) -> Optional[Tuple[int, int]]:
        if label := key.get_label(self.LAYOUT_OPTION_LABEL):
            parts = label.split(",")
            if len(parts) != 2:
                msg = "Unexpected number of ',' delimited elements in key label"
                raise ValueError(msg)
            return int(parts[0]), int(parts[1])
        return None

    def _get_layout_options(self) -> Dict[int, Dict[int, List[Key]]]:
        keys: Dict[int, Dict[int, List[Key]]] = defaultdict(lambda: defaultdict(list))
        for key in self.key_iterator(ignore_alternative=False):
            option = self._get_layout_option_or_none(key)
            if option:
                keys[option[0]][option[1]].append(key)
        return keys

    @staticmethod
    def _key_matrix_position(key: Key) -> Tuple[int, int, int]:
        matrix_position = MatrixAnnotatedKeyboard.get_matrix_position(key)
        row_match = re.search(r"\d+", matrix_position[0])
        column_match = re.search(r"\d+", matrix_position[1])

        if row_match is None or column_match is None:
            msg = f"No numeric part for row or column found in '{matrix_position}'"
            raise ValueError(msg)

        return (
            int(row_match.group()),
            int(column_match.group()),
            MatrixAnnotatedKeyboard.get_layout_option(key),
        )

    def collapse(self) -> None:
        """Modify positions of alternative_keys to destination positions
        i.e. the positions they would take as 'non alternative'
        and de-duplicate items with equal matrix coordinates and size
        """
        if self.collapsed:
            # prevent double collapsing
            return

        seen = {}
        new_alternatives = []

        def _int(value: float) -> Union[int, float]:
            return int(value) if int(value) == value else value

        def _key_center(key: Key) -> Tuple[Union[int, float], Union[int, float]]:
            return (_int(key.x + key.width / 2), _int(key.y + key.height / 2))

        def _key_props(key: Key):
            props = (
                key.labels[self.MATRIX_COORDINATES_LABEL],
                *_key_center(key),
                key.decal,
            )
            return props

        for k in self.keys:
            # ignore decals in default key group
            if k.decal:
                continue
            seen[_key_props(k)] = True

        layout_keys = self._get_layout_options()
        for choices in layout_keys.values():
            anchor = min(choices[0], key=lambda key: (key.x, key.y))
            for choice, keys in choices.items():
                if choice != 0:
                    group_anchor = min(keys, key=lambda key: (key.x, key.y))
                    move_x = anchor.x - group_anchor.x
                    move_y = anchor.y - group_anchor.y
                    for k in keys:
                        k.x = _int(k.x + move_x)
                        k.y = _int(k.y + move_y)
                        props = _key_props(k)
                        if props not in seen:
                            seen[props] = True
                            new_alternatives.append(k)

        for key in list(new_alternatives):
            if key.decal:
                new_alternatives.remove(key)

        self.alternative_keys = new_alternatives
        self.collapsed = True

    def sort_keys(self) -> None:
        for l in [self.keys, self.alternative_keys]:
            l.sort(key=lambda k: MatrixAnnotatedKeyboard._key_matrix_position(k))

    def keys_in_matrix_order(self) -> List[Key]:
        """Returns keys in matrix row/column order. If multiple keys occupy same
        matrix position, sort by layout option label. Ignores decal keys.
        """
        items: List[Key] = []
        for key in self.key_iterator(ignore_alternative=False):
            if key.decal:
                continue
            items.append(key)

        return sorted(
            items, key=lambda k: MatrixAnnotatedKeyboard._key_matrix_position(k)
        )

    @staticmethod
    def get_matrix_position(key: Key) -> Tuple[str, str]:
        try:
            label = key.get_label(MatrixAnnotatedKeyboard.MATRIX_COORDINATES_LABEL)
            split = str(label).split(",")
            if len(split) != 2:
                raise RuntimeError
            return (split[0].strip(), split[1].strip())
        except Exception as e:
            msg = "Matrix coordinates label missing or invalid"
            raise RuntimeError(msg) from e

    @staticmethod
    def get_layout_option(key: Key) -> int:
        if layout_option_label := key.get_label(
            MatrixAnnotatedKeyboard.LAYOUT_OPTION_LABEL
        ):
            return int(layout_option_label.split(",")[1])
        return 0

    def to_keyboard(self) -> Keyboard:
        return Keyboard(meta=self.meta, keys=self.keys_in_matrix_order())


def reorder_items(items: List[Any], align: int) -> List[Any]:
    ret: List[Any] = KEY_MAX_LABELS * [None]
    for i, item in enumerate(items):
        if item:
            index = LABEL_MAP[align][i]
            ret[index] = item
    while ret and ret[-1] is None:
        ret.pop()
    return ret


def reorder_items_kle(items, align) -> List[Any]:
    ret: List[Any] = KEY_MAX_LABELS * [None]
    for i, label in enumerate(items):
        if label:
            index = REVERSE_LABEL_MAP[align][i]
            if index == -1:
                ret = []
                break
            ret[index] = label
    while ret and ret[-1] is None:
        ret.pop()
    return ret


def find_best_label_alignment(labels) -> Tuple[int, List[Any]]:
    results = {}
    for align in reversed(range(0, 8)):
        if ret := reorder_items_kle(labels, align):
            results[align] = ret

    if results.items():
        best = min(results.items(), key=lambda x: len(x[1]))
        return best[0], best[1]
    else:
        return 0, []


def cleanup_key(key: Key) -> None:
    for attribute_name in ["textSize", "textColor"]:
        attribute = getattr(key, attribute_name)
        attribute = attribute[0 : len(key.labels)]
        for i, (label, val) in enumerate(zip(key.labels, attribute)):
            if not label:
                attribute[i] = None
            if val == getattr(key.default, attribute_name):
                attribute[i] = None
        while attribute and attribute[-1] is None:
            attribute.pop()
        setattr(key, attribute_name, attribute)


def parse_qmk(layout) -> MatrixAnnotatedKeyboard:
    metadata: KeyboardMetadata = KeyboardMetadata()

    layouts = layout["layouts"]
    keys: Dict[Tuple[int, int], List[Key]] = defaultdict(list)

    for i, layout in enumerate(layouts.values()):
        for item in layout["layout"]:
            if not isinstance(item, dict):
                msg = f"Unexpected data appeared while parsing QMK layout: '{item}'"
                raise RuntimeError(msg)

            key: Key = Key()
            key.x = item["x"]
            key.y = item["y"]
            if "r" in item:
                key.rotation_angle = item["r"]
            if "rx" in item:
                key.rotation_x = item["rx"]
            if "ry" in item:
                key.rotation_y = item["ry"]
            if "w" in item:
                key.width = item["w"]
            if "h" in item:
                key.height = item["h"]

            # qmk uses 'matrix' property to define key position instead of labels
            # this is not a part of Key dataclass derived from
            # keyboard-layout-editor schema. Because we do not care about actual
            # labels (which qmk layout also can define), use approach from via
            # layouts, i.e. encode matrix position in first label
            matrix_position = item["matrix"]
            if not isinstance(matrix_position, list):
                msg = (
                    "Unexpected key matrix position appeared while parsing QMK "
                    f"layout: '{matrix_position}'"
                )
                raise RuntimeError(msg)
            key.set_label(
                MatrixAnnotatedKeyboard.MATRIX_COORDINATES_LABEL,
                f"{matrix_position[0]},{matrix_position[1]}",
            )
            # qmk layouts do not have information about alternative layout option groups,
            # each group is 0
            key.set_label(MatrixAnnotatedKeyboard.LAYOUT_OPTION_LABEL, f"0,{i}")

            position = (matrix_position[0], matrix_position[1])
            keys[position].append(key)

    # remove duplicate keys ignoring LAYOUT_OPTION_LABEL value
    deduplicate_keys: Dict[Tuple[int, int], List[Key]] = defaultdict(list)
    for position, key_list in keys.items():
        deduplicate_position_keys: List[Key] = []
        for k in key_list:
            duplicate = False
            for k1 in deduplicate_position_keys:
                temp1 = copy.deepcopy(k)
                temp2 = copy.deepcopy(k1)
                temp1.set_label(MatrixAnnotatedKeyboard.LAYOUT_OPTION_LABEL, "")
                temp2.set_label(MatrixAnnotatedKeyboard.LAYOUT_OPTION_LABEL, "")
                if temp1 == temp2:
                    duplicate = True
                    break
            if not duplicate:
                deduplicate_position_keys.append(k)

        # clean up labels, i.e. remove LAYOUT_OPTION_LABEL if given key does not have
        # alternative value
        if len(deduplicate_position_keys) == 1:
            deduplicate_position_keys[0].labels = [
                deduplicate_position_keys[0].labels[0]
            ]

        deduplicate_keys[position] = deduplicate_position_keys

    final_keys: List[Key] = list()
    for _, l in iter(sorted(deduplicate_keys.items())):
        final_keys += l

    keyboard = MatrixAnnotatedKeyboard(meta=metadata, keys=final_keys)
    keyboard.collapsed = True
    return keyboard


def parse_via(layout) -> MatrixAnnotatedKeyboard:
    keyboard = parse_kle(layout["layouts"]["keymap"])
    return MatrixAnnotatedKeyboard(meta=keyboard.meta, keys=keyboard.keys)


def parse_kle(layout) -> Keyboard:
    if not isinstance(layout, list):
        msg = "Expected an list of objects"
        raise RuntimeError(msg)

    metadata: KeyboardMetadata = KeyboardMetadata()
    rows: List[Any] = layout
    current: Key = Key()

    if not rows:
        msg = "Expected at least one row of keys"
        raise RuntimeError(msg)

    keys = []
    cluster = {"x": 0, "y": 0}
    align = 4

    for r, row in enumerate(rows):
        if isinstance(row, list):
            for k, item in enumerate(row):
                if isinstance(item, str):
                    new_key = copy.deepcopy(current)
                    # Calculate some generated values
                    new_key.width2 = (
                        current.width if new_key.width2 == 0 else current.width2
                    )
                    new_key.height2 = (
                        current.height if new_key.height2 == 0 else current.height2
                    )
                    items = item.split("\n")
                    if len(items) > KEY_MAX_LABELS:
                        msg = (
                            f"Illegal key labels: '{repr(item)}'. "
                            f"Labels string can contain {KEY_MAX_LABELS} '\\n' "
                            "separated items, ignoring redundant values."
                        )
                        logger.warning(msg)
                        items = items[0:KEY_MAX_LABELS]
                    new_key.labels = reorder_items(items, align)
                    new_key.textSize = reorder_items(new_key.textSize, align)

                    cleanup_key(new_key)

                    keys.append(new_key)

                    current.x = round(current.x + current.width, 6)
                    current.width = 1
                    current.height = 1
                    current.x2 = 0
                    current.y2 = 0
                    current.width2 = 0
                    current.height2 = 0
                    current.nub = False
                    current.stepped = False
                    current.decal = False
                elif isinstance(item, dict):
                    if k != 0 and ("r" in item or "rx" in item or "ry" in item):
                        msg = (
                            "Rotation can only be specified on the first key in the row"
                        )
                        raise RuntimeError(msg)
                    if "r" in item:
                        current.rotation_angle = item["r"]
                    if "rx" in item:
                        cluster["x"] = item["rx"]
                        current.x = cluster["x"]
                        current.y = cluster["y"]
                        current.rotation_x = item["rx"]
                    if "ry" in item:
                        cluster["y"] = item["ry"]
                        current.x = cluster["x"]
                        current.y = cluster["y"]
                        current.rotation_y = item["ry"]
                    if "a" in item:
                        align = item["a"]
                    if "f" in item:
                        current.default.textSize = item["f"]
                        current.textSize = []
                    if "f2" in item:
                        if len(current.textSize) == 0:
                            current.textSize = [None]
                        for _ in range(1, KEY_MAX_LABELS):
                            current.textSize.append(item["f2"])
                    if "fa" in item:
                        current.textSize = item["fa"]
                    if "p" in item:
                        current.profile = item["p"]
                    if "c" in item:
                        current.color = item["c"]
                    if "t" in item:
                        split = item["t"].split("\n")
                        if split[0]:
                            current.default.textColor = split[0]
                        current.textColor = reorder_items(split, align)
                    if "x" in item:
                        current.x = round(current.x + item["x"], 6)
                    if "y" in item:
                        current.y = round(current.y + item["y"], 6)
                    if "w" in item:
                        current.width = item["w"]
                        current.width2 = item["w"]
                    if "h" in item:
                        current.height = item["h"]
                        current.height2 = item["h"]
                    if "x2" in item:
                        current.x2 = item["x2"]
                    if "y2" in item:
                        current.y2 = item["y2"]
                    if "w2" in item:
                        current.width2 = item["w2"]
                    if "h2" in item:
                        current.height2 = item["h2"]
                    if "n" in item:
                        current.nub = item["n"]
                    if "l" in item:
                        current.stepped = item["l"]
                    if "d" in item:
                        current.decal = item["d"]
                    if "g" in item:
                        current.ghost = item["g"]
                    if "sm" in item:
                        current.sm = item["sm"]
                    if "sb" in item:
                        current.sb = item["sb"]
                    if "st" in item:
                        current.st = item["st"]
                else:
                    msg = "Unexpected item type"
                    raise RuntimeError(msg)

            # end of the row:
            current.y = round(current.y + 1, 6)
            current.x = current.rotation_x
        elif isinstance(row, dict) and r == 0:
            field_set = {f.name for f in fields(KeyboardMetadata) if f.init}
            row_filtered = {k: v for k, v in row.items() if k in field_set}
            metadata = KeyboardMetadata(**row_filtered)
        else:
            msg = "Unexpected"
            raise RuntimeError(msg)

    return Keyboard(meta=metadata, keys=keys)


def parse_ergogen_points(layout: dict, *, zone_filter: str = "") -> Keyboard:
    if not isinstance(layout, dict):
        msg = "Expected an object with nested objects"
        raise RuntimeError(msg)
    if not layout:
        msg = "Expected non-empty object"
        raise RuntimeError(msg)

    metadata: KeyboardMetadata = KeyboardMetadata()
    keys = []

    # in ergogen terminology, 'padding' is vertical space between each key
    # and 'spread' is horizontal space between each key. These values are defined
    # in 'meta' object for each 'point' (key). Let's assume that these values are
    # the same for all of them (which I don't know if it is actually true)
    padding = None
    spread = None

    # at the end of parsing will need to fix key order, kle starts from top,
    # ergogen from bottom
    # store topmost, leftmost key index while iterating, it will be used later
    topmost_leftmost = [0, 0]

    def __int(value: float) -> Union[int, float]:
        return int(value) if int(value) == value else value

    pattern = re.compile(zone_filter) if zone_filter else None

    for name, item in layout.items():
        if pattern and not re.match(pattern, name):
            continue
        if "meta" not in item:
            msg = "Item needs to have meta defined"
            raise RuntimeError(msg)

        meta = item["meta"]

        # read key spacing only once, assume that it does not change
        if not padding and not spread:
            if "padding" not in meta or "spread" not in meta:
                msg = "Unable to determine key spacing"
                raise RuntimeError(msg)
            padding = meta["padding"]
            spread = meta["spread"]

        key = Key()

        # kle expresses distances (positions) with 1U, need to normalize
        # ergogen uses key center for position, kle uses key left top corner,
        # need to adjust position by size
        key.x = __int(item["x"] / spread)
        key.y = __int(item["y"] / padding)
        key.width = __int(meta["width"] / spread)
        key.height = __int(meta["height"] / spread)

        # non-custom key shapes (like ISO enter) not supported:
        key.width2 = key.width
        key.height2 = key.height

        if key.y > topmost_leftmost[1] or (
            key.y == topmost_leftmost[1] and key.x <= topmost_leftmost[0]
        ):
            topmost_leftmost = [key.x, key.y]

        # kle and ergogen rotate in opposite directions
        key.rotation_angle = -1 * item["r"]

        # if column_net and row_net defined, add it to label
        row = meta.get("row_net", "")
        column = meta.get("column_net", "")
        if row and column:
            key.labels.append(f"{row},{column}")

        keys.append(key)

    # do some cleanup to be kle compatible
    # adjust coordinates
    for key in keys:
        # reverse top-bottom
        key.y = abs(key.y - topmost_leftmost[1])
        # move position from key center (ergogen) to top left corner (kle)
        key.x = key.x - key.width / 2
        key.y = key.y - key.height / 2
    # move out of negative positions
    min_x = min(keys, key=lambda k: k.x).x
    if min_x >= 0:
        min_x = 0

    min_y = min(keys, key=lambda k: k.y).y
    if min_y >= 0:
        min_y = 0

    for key in keys:
        key.x = key.x - min_x
        key.y = key.y - min_y
        # rotation is always expressed in relation to key center
        key.rotation_x = key.x + key.width / 2 if key.rotation_angle else 0
        key.rotation_y = key.y + key.height / 2 if key.rotation_angle else 0

    # and sort (topmost leftmost first)
    keys = sorted(keys, key=lambda k: [k.y, k.x])

    return Keyboard(meta=metadata, keys=keys)


def get_keyboard(layout: dict) -> Keyboard:
    try:
        return parse_kle(layout)
    except Exception:
        pass
    try:
        return parse_via(layout)
    except Exception:
        pass
    try:
        return parse_qmk(layout)
    except Exception:
        pass
    try:
        return Keyboard.from_json(layout)
    except Exception:
        pass
    try:
        return parse_ergogen_points(layout)
    except Exception:
        pass
    msg = "Unable to get keyboard layout"
    raise RuntimeError(msg)


def get_keyboard_from_file(layout_path: str) -> Keyboard:
    # Layout downloaded from keyboard-layout-editor is most likely using utf-8.
    # Use it explicitly in case the platform locale sets different encoding.
    with open(layout_path, "r", encoding="utf-8") as f:
        layout = json.load(f)
    logger.info(f"User layout: {layout}")
    return get_keyboard(layout)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="KLE format converter")
    parser.add_argument("-in", required=True, help="Layout file")
    parser.add_argument(
        "-inform",
        required=False,
        default="KLE_RAW",
        choices=["KLE_RAW", "KLE_VIA", "KLE_INTERNAL", "ERGOGEN_INTERNAL", "QMK"],
        help="Specifies the input format",
    )
    parser.add_argument("-out", required=False, help="Result file")
    parser.add_argument(
        "-outform",
        required=False,
        default="KLE_INTERNAL",
        choices=["KLE_RAW", "KLE_INTERNAL"],
        help="Specifies the output format",
    )
    parser.add_argument(
        "-text", required=False, action="store_true", help="Print result"
    )
    parser.add_argument(
        "-ergogen-filter",
        required=False,
        type=str,
        help="Ergogen zone filter regular expression, applicable only when -inform ERGOGEN_INTERNAL",
    )
    parser.add_argument(
        "-collapse",
        action="store_true",
        help=(
            "Collapse via-like annotated layout, "
            "applicable only when -inform equal KLE_RAW, KLE_VIA or KLE_INTERNAL"
        ),
    )

    args = parser.parse_args()
    input_path = getattr(args, "in")
    input_format = args.inform
    output_path = getattr(args, "out")
    output_format = args.outform
    print_result = args.text
    ergogen_filter = args.ergogen_filter
    collapse = args.collapse

    if input_format == output_format and not collapse:
        print("Output format equal input format, nothing to do...")
        sys.exit(1)

    def _keyboard_to_kle_raw(keyboard: Keyboard):
        result = keyboard.to_kle()
        if print_result:
            print(result)
        # 'to_kle' returns 'raw data' string which can be copy pasted
        # to keyboard-layout-editor, to make json out of it we need
        # to wrap it in list. Then it can be uploaded as JSON.
        result = "[" + result + "]"
        return json.loads(result)

    def _keyboard_to_kle_internal(keyboard: Keyboard):
        result = json.loads(keyboard.to_json())
        if print_result:
            pprint.pprint(result)
        return result

    with open(input_path, "r", encoding="utf-8") as input_file:
        if input_path.endswith("yaml") or input_path.endswith("yml"):
            try:
                import yaml

                layout = yaml.safe_load(input_file)
            except Exception as e:
                msg = (
                    "Could not load yaml file, make sure that `PyYAML` installed "
                    "and yaml file format correct"
                )
                raise RuntimeError(msg) from e
        else:
            layout = json.load(input_file)

        result = ""
        if input_format == "KLE_RAW":
            keyboard = parse_kle(layout)
        elif input_format == "KLE_VIA":
            # 'parse_via' creates MatrixAnnotatedKeyboard which is our
            # internal representation and it is not the same thing
            # as KLE_INTERNAL format, so it is not used here
            keyboard = parse_kle(layout["layouts"]["keymap"])
        elif input_format == "KLE_INTERNAL":
            keyboard = Keyboard.from_json(layout)
        elif input_format == "ERGOGEN_INTERNAL":
            keyboard = parse_ergogen_points(layout, zone_filter=ergogen_filter)
        else:  # QMK
            keyboard = parse_qmk(layout).to_keyboard()

        if collapse:
            keyboard = MatrixAnnotatedKeyboard(meta=keyboard.meta, keys=keyboard.keys)
            keyboard.collapse()
            keyboard = keyboard.to_keyboard()

        if output_format == "KLE_INTERNAL":
            result = _keyboard_to_kle_internal(keyboard)
        else:  # KLE_RAW
            result = _keyboard_to_kle_raw(keyboard)

        if output_path:
            with open(output_path, "w", encoding="utf-8") as output_file:
                json.dump(result, output_file, indent=2)



from __future__ import annotations

import gettext
import json
import logging
import os
import string
import sys
from dataclasses import asdict, dataclass, field
from enum import Flag
from typing import List, Optional, Tuple

import wx
from wx.lib.embeddedimage import PyEmbeddedImage

from .defaults import DEFAULT_DIODE_POSITION, ZERO_POSITION
from .element_position import ElementInfo, ElementPosition, PositionOption, Side
from .help_dialog import HelpDialog

logger = logging.getLogger(__name__)
TEXT_CTRL_EXTRA_SPACE = 25

# Most of the phrases used in this plugin are already in use in KiCad.
# It means that we get translations for free using `wx.GetTranslation`.
# All strings translated with 'wx_' are expected to be a part of
# KiCad's translation files. All remaining will be translated with
# another custom mechanism or will remain default.
wx_ = wx.GetTranslation

# Currently there is no elegant way to check which language is loaded by KiCad.
# This feature has been requested here:
#   https://gitlab.com/kicad/code/kicad/-/issues/10573
# Until then, use workaroud - request translation with wx_ and use result
# in lookup table. This lookup should contain all installed languages defined
# in translation/pofiles/LINGUAS_INSTALL.
KICAD_TRANSLATIONS_LOOKUP = {
    "Sprache": "de",
    "Set Language": "en",
    "Seleccionar idioma": "es",
    "言語設定": "ja",
    "언어 설정": "ko",
    "Taal instellen": "nl",
    "Ustaw język": "pl",
    "Установить язык": "ru",
    "Nastaviť jazyk": "sk",
    "Встановити мову": "uk",
    "設定語言": "zh_CN",
}


@dataclass
class WindowState:
    layout_path: str = ""
    key_distance: Tuple[float, float] = (19.05, 19.05)
    key_info: ElementInfo = field(
        default_factory=lambda: ElementInfo(
            "SW{}", PositionOption.DEFAULT, ZERO_POSITION, ""
        )
    )
    enable_diode_placement: bool = True
    route_switches_with_diodes: bool = True
    optimize_diodes_orientation: bool = False
    diode_info: ElementInfo = field(
        default_factory=lambda: ElementInfo(
            "D{}", PositionOption.DEFAULT, DEFAULT_DIODE_POSITION, ""
        )
    )
    additional_elements: List[ElementInfo] = field(
        default_factory=lambda: [
            ElementInfo("ST{}", PositionOption.CUSTOM, ZERO_POSITION, "")
        ]
    )
    route_rows_and_columns: bool = True
    template_path: str = ""
    generate_outline: bool = False
    outline_delta: float = 0.0

    def __str__(self) -> str:
        return json.dumps(asdict(self), indent=None)

    @classmethod
    def from_dict(cls, data: dict) -> WindowState:
        key_distance = data.pop("key_distance")
        if type(key_distance) is list:
            key_distance = tuple(key_distance)
        key_info = ElementInfo.from_dict(data.pop("key_info"))
        diode_info = ElementInfo.from_dict(data.pop("diode_info"))
        additional_elements = [
            ElementInfo.from_dict(i) for i in data.pop("additional_elements")
        ]
        return cls(
            key_distance=key_distance,
            key_info=key_info,
            diode_info=diode_info,
            additional_elements=additional_elements,
            **data,
        )


def get_current_kicad_language() -> str:
    kicad_lang = KICAD_TRANSLATIONS_LOOKUP.get(wx_("Set Language"), "en")
    # some languages require additional lookup to detect correctly,
    # for example es vs es_MX:
    if kicad_lang == "es" and wx_("Enabled:") == "Activado:":
        # es: Enabled -> Habilitado, es_MX: Enabled -> Activado
        kicad_lang = "es_MX"
    return kicad_lang


def get_plugin_translator(lang: str = "en"):
    localedir = os.path.join(os.path.dirname(__file__), "locale")
    trans = gettext.translation(
        "kbplacer", localedir=localedir, languages=(lang,), fallback=True
    )
    return trans.gettext


def get_file_picker(*args, **kwargs) -> wx.FilePickerCtrl:
    file_picker = wx.FilePickerCtrl(*args, **kwargs)
    file_picker.SetTextCtrlGrowable(True)

    def _update_position(_) -> None:
        text_ctrl = file_picker.GetTextCtrl()
        # when updating from file chooser window we want automatically
        # move to the end (it looks better when not whole path fits),
        # but when user updates by typing in text control we don't
        # want to mess up with it. This can be done by checking
        # if current insertion point is 0.
        if text_ctrl.GetInsertionPoint() == 0:
            text_ctrl.SetInsertionPointEnd()

    file_picker.Bind(wx.EVT_FILEPICKER_CHANGED, _update_position)
    return file_picker


class FloatValidator(wx.Validator):
    def __init__(self) -> None:
        wx.Validator.__init__(self)
        self.Bind(wx.EVT_CHAR, self.OnChar)

    def Clone(self) -> FloatValidator:
        return FloatValidator()

    def Validate(self, _) -> bool:
        text_ctrl = self.GetWindow()
        if not text_ctrl.IsEnabled():
            return True

        text = text_ctrl.GetValue()
        try:
            float(text)
            return True
        except ValueError:
            # this can happen when value is empty, equal '-', '.', or '-.',
            # other invalid values should not be allowed by 'OnChar' filtering
            name = text_ctrl.GetName()
            wx.MessageBox(f"Invalid '{name}' value: '{text}' is not a number!", "Error")
            text_ctrl.SetFocus()
            return False

    def TransferToWindow(self) -> bool:
        return True

    def TransferFromWindow(self) -> bool:
        return True

    def OnChar(self, event: wx.KeyEvent) -> None:
        text_ctrl = self.GetWindow()
        current_position = text_ctrl.GetInsertionPoint()
        keycode = int(event.GetKeyCode())
        if keycode in [
            wx.WXK_BACK,
            wx.WXK_DELETE,
            wx.WXK_LEFT,
            wx.WXK_RIGHT,
            wx.WXK_NUMPAD_LEFT,
            wx.WXK_NUMPAD_RIGHT,
            wx.WXK_TAB,
        ]:
            event.Skip()
        else:
            text_ctrl = self.GetWindow()
            text = text_ctrl.GetValue()
            key = chr(keycode)
            if (
                # allow only digits
                # or single '-' when as first character
                # or single '.'
                key in string.digits
                or (key == "-" and "-" not in text and current_position == 0)
                or (key == "." and "." not in text)
            ):
                event.Skip()


class LabeledTextCtrl(wx.Panel):
    def __init__(
        self,
        parent: wx.Window,
        label: str,
        value: str,
        width: int = -1,
        validator: wx.Validator = wx.DefaultValidator,
    ) -> None:
        super().__init__(parent)

        expected_char_width = self.GetTextExtent("x").x
        if width != -1:
            annotation_format_size = wx.Size(
                expected_char_width * width + TEXT_CTRL_EXTRA_SPACE, -1
            )
        else:
            annotation_format_size = wx.Size(-1, -1)

        self.label = wx.StaticText(self, -1, label)
        self.text = wx.TextCtrl(
            self,
            value=value,
            size=annotation_format_size,
            validator=validator,
            name=label.strip(":"),
        )

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(self.label, 0, wx.RIGHT | wx.ALIGN_CENTER_VERTICAL, 5)
        sizer.Add(self.text, 0, wx.EXPAND | wx.TOP | wx.BOTTOM, 5)

        self.SetSizer(sizer)

    def Enable(self) -> None:
        self.label.Enable()
        self.text.Enable()

    def Disable(self) -> None:
        self.label.Disable()
        self.text.Disable()

    def Hide(self) -> None:
        self.label.Hide()
        self.text.Hide()


class CustomRadioBox(wx.Panel):
    def __init__(self, parent: wx.Window, choices: List[str]) -> None:
        super().__init__(parent)
        self.radio_buttons: dict[str, wx.RadioButton] = {}

        for choice in choices:
            radio_button = wx.RadioButton(self, label=choice)
            self.radio_buttons[choice] = radio_button

        # this is special hidden option to allow clearing (selecting none)
        self.none_button = wx.RadioButton(self, label="")
        self.none_button.Hide()

        # on linux, use negative border because otherwise it looks too spaced out
        space = 0 if sys.platform == "win32" else -2
        sizer = wx.BoxSizer(wx.VERTICAL)
        for radio_button in self.radio_buttons.values():
            sizer.Add(radio_button, 0, wx.TOP | wx.BOTTOM, space)

        self.SetSizer(sizer)

    def Select(self, choice: str) -> None:
        self.radio_buttons[choice].SetValue(True)

    def Clear(self) -> None:
        self.none_button.SetValue(True)

    def GetValue(self) -> Optional[str]:
        if not self.none_button.GetValue():
            for choice, button in self.radio_buttons.items():
                if button.GetValue():
                    return choice
        return None


class ElementPositionWidget(wx.Panel):
    def __init__(
        self,
        parent: wx.Window,
        default_position: Optional[ElementPosition] = None,
        disable_offsets: bool = False,
    ) -> None:
        super().__init__(parent)

        self.default = default_position
        self.x = LabeledTextCtrl(
            self, wx_("Offset X:"), value="", width=5, validator=FloatValidator()
        )
        self.y = LabeledTextCtrl(
            self, wx_("Y:"), value="", width=5, validator=FloatValidator()
        )
        self.orientation = LabeledTextCtrl(
            self, wx_("Orientation:"), value="", width=5, validator=FloatValidator()
        )
        if disable_offsets:
            self.x.Hide()
            self.y.Hide()
        self.side_label = wx.StaticText(self, -1, wx_("Side:"))
        self.side = CustomRadioBox(self, choices=[wx_("Front"), wx_("Back")])

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        if not disable_offsets:
            sizer.Add(self.x, 0, wx.EXPAND | wx.LEFT, 5)
            sizer.Add(self.y, 0, wx.EXPAND | wx.LEFT, 5)
        sizer.Add(self.orientation, 0, wx.EXPAND | wx.LEFT, 5)
        sizer.Add(self.side_label, 0, wx.LEFT | wx.ALIGN_CENTER_VERTICAL, 5)
        sizer.Add(self.side, 0, wx.EXPAND | wx.LEFT, 5)

        self.SetSizer(sizer)

    def set_position_by_choice(self, choice: str) -> None:
        if choice == PositionOption.DEFAULT:
            self.__set_position_to_default()
        elif choice == PositionOption.CUSTOM:
            self.__set_position_to_zero_editable()
        else:
            self.__set_position_to_empty_non_editable()

    def set_position(self, position: ElementPosition) -> None:
        self.__set_coordinates(str(position.x), str(position.y))
        self.__set_orientation(str(position.orientation))
        self.__set_side(position.side)

    def __set_position_to_default(self) -> None:
        if self.default:
            x = str(self.default.x)
            y = str(self.default.y)
            self.__set_coordinates(x, y)
            self.__set_orientation(str(self.default.orientation))
            self.__set_side(self.default.side)
            self.Disable()

    def __set_position_to_zero_editable(self) -> None:
        self.__set_coordinates("0", "0")
        self.__set_orientation("0")
        self.__set_side(Side.FRONT)
        self.Enable()

    def __set_position_to_empty_non_editable(self) -> None:
        self.__set_coordinates("-", "-")
        self.__set_orientation("-")
        self.side.Clear()
        self.Disable()

    def __set_coordinates(self, x: str, y: str) -> None:
        self.x.text.SetValue(x)
        self.y.text.SetValue(y)

    def __set_orientation(self, orientation: str) -> None:
        self.orientation.text.SetValue(orientation)

    def __set_side(self, side: Side) -> None:
        if side == Side.BACK:
            self.side.Select(wx_("Back"))
        else:
            self.side.Select(wx_("Front"))

    def __get_side(self) -> Side:
        side_str = str(self.side.GetValue())
        if side_str == wx_("Back"):
            return Side.BACK
        else:
            return Side.FRONT

    def GetValue(self) -> ElementPosition:
        x = float(self.x.text.GetValue())
        y = float(self.y.text.GetValue())
        orientation = float(self.orientation.text.GetValue())
        return ElementPosition(x, y, orientation, self.__get_side())

    def Enable(self) -> None:
        self.x.Enable()
        self.y.Enable()
        self.orientation.Enable()
        self.side_label.Enable()
        self.side.Enable()

    def Disable(self) -> None:
        self.x.Disable()
        self.y.Disable()
        self.orientation.Disable()
        self.side_label.Disable()
        self.side.Disable()


class TemplateType(Flag):
    LOAD = False
    SAVE = True


class ElementTemplateSelectionWidget(wx.Panel):
    def __init__(
        self, parent: wx.Window, picker_type: TemplateType, initial_path: str = ""
    ) -> None:
        super().__init__(parent)
        self._ = self.GetTopLevelParent()._

        label = (
            self._("Save to:")
            if picker_type == TemplateType.SAVE
            else self._("Load from:")
        )
        layout_label = wx.StaticText(self, -1, label)
        layout_picker = get_file_picker(
            self,
            -1,
            wildcard="KiCad printed circuit board files (*.kicad_pcb)|*.kicad_pcb",
            style=wx.FLP_USE_TEXTCTRL
            | (wx.FLP_SAVE if picker_type == TemplateType.SAVE else wx.FLP_OPEN),
        )

        if initial_path:
            layout_picker.SetPath(initial_path)
            layout_picker.GetTextCtrl().SetInsertionPointEnd()

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(layout_label, 0, wx.LEFT | wx.RIGHT | wx.ALIGN_CENTER_VERTICAL, 5)
        sizer.Add(layout_picker, 1, wx.EXPAND | wx.ALL, 5)
        self.SetSizer(sizer)

        self.__layout_picker = layout_picker

    def GetValue(self) -> str:
        return self.__layout_picker.GetPath()


class ElementPositionChoiceWidget(wx.Panel):
    def __init__(
        self,
        parent: wx.Window,
        initial_choice: PositionOption,
        initial_position: Optional[ElementPosition] = None,
        default_position: Optional[ElementPosition] = None,
        initial_path: str = "",
    ) -> None:
        super().__init__(parent)

        choices = [
            PositionOption.CUSTOM,
            PositionOption.RELATIVE,
            PositionOption.PRESET,
        ]
        if default_position:
            choices.insert(0, PositionOption.DEFAULT)

        self.dropdown = wx.ComboBox(self, choices=choices, style=wx.CB_DROPDOWN)
        self.dropdown.Bind(wx.EVT_COMBOBOX, self.__on_position_choice_change)

        dropdown_sizer = wx.BoxSizer(wx.HORIZONTAL)
        dropdown_sizer.Add(
            wx.StaticText(self, -1, wx_("Position") + ":"),
            0,
            wx.RIGHT | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        dropdown_sizer.Add(self.dropdown, 0, wx.EXPAND | wx.TOP | wx.BOTTOM, 5)

        self.position = ElementPositionWidget(self, default_position)
        self.load_template = ElementTemplateSelectionWidget(
            self, picker_type=TemplateType.LOAD, initial_path=initial_path
        )
        self.save_template = ElementTemplateSelectionWidget(
            self, picker_type=TemplateType.SAVE, initial_path=initial_path
        )

        self.__set_initial_state(initial_choice)
        if initial_position and initial_choice == PositionOption.CUSTOM:
            self.position.set_position(initial_position)

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(dropdown_sizer, 0, wx.EXPAND | wx.ALL, 5)
        sizer.Add(self.position, 0, wx.EXPAND | wx.ALL, 5)
        sizer.Add(self.load_template, 1, wx.EXPAND | wx.ALL, 5)
        sizer.Add(self.save_template, 1, wx.EXPAND | wx.ALL, 5)

        self.SetSizer(sizer)

    def __set_initial_state(self, choice: str) -> None:
        self.dropdown.SetValue(choice)
        self.__set_position_by_choice(choice)

    def __on_position_choice_change(self, event: wx.CommandEvent) -> None:
        choice = event.GetString()
        self.__set_position_by_choice(choice)

    def __set_position_by_choice(self, choice: str) -> None:
        if choice == PositionOption.RELATIVE:
            self.position.Hide()
            self.load_template.Hide()
            self.save_template.Show()
        elif choice == PositionOption.PRESET:
            self.position.Hide()
            self.load_template.Show()
            self.save_template.Hide()
        else:
            self.position.Show()
            self.load_template.Hide()
            self.save_template.Hide()
        self.GetTopLevelParent().Layout()
        self.position.set_position_by_choice(choice)
        self.choice = PositionOption(choice)

    def GetValue(self) -> Tuple[PositionOption, Optional[ElementPosition], str]:
        template_path = ""
        if self.dropdown.GetValue() == PositionOption.RELATIVE:
            template_path = self.save_template.GetValue()
        elif self.dropdown.GetValue() == PositionOption.PRESET:
            template_path = self.load_template.GetValue()

        if self.choice not in [PositionOption.DEFAULT, PositionOption.CUSTOM]:
            return self.choice, None, template_path
        return self.choice, self.position.GetValue(), template_path

    def Enable(self) -> None:
        self.dropdown.Enable()
        if self.dropdown.GetValue() == PositionOption.CUSTOM:
            self.position.Enable()
        self.load_template.Enable()
        self.save_template.Enable()

    def Disable(self) -> None:
        self.dropdown.Disable()
        self.position.Disable()
        self.load_template.Disable()
        self.save_template.Disable()


class ElementSettingsWidget(wx.Panel):
    def __init__(
        self,
        parent: wx.Window,
        element_info: ElementInfo,
        default_position: Optional[ElementPosition] = None,
    ) -> None:
        super().__init__(parent)

        self.annotation_format = LabeledTextCtrl(
            self,
            label=wx_("Footprint Annotation") + ":",
            value=element_info.annotation_format,
            width=3,
        )
        self.position_widget = ElementPositionChoiceWidget(
            self,
            element_info.position_option,
            element_info.position,
            default_position=default_position,
            initial_path=element_info.template_path,
        )

        sizer = wx.BoxSizer(wx.HORIZONTAL)

        sizer.Add(
            self.annotation_format, 0, wx.EXPAND | wx.TOP | wx.BOTTOM | wx.RIGHT, 5
        )
        sizer.Add(self.position_widget, 1, wx.EXPAND | wx.ALL, 0)

        self.SetSizer(sizer)

    def GetValue(self) -> ElementInfo:
        annotation = self.annotation_format.text.GetValue()
        position = self.position_widget.GetValue()
        return ElementInfo(annotation, *position)

    def Enable(self) -> None:
        self.position_widget.Enable()

    def Disable(self) -> None:
        self.position_widget.Disable()


class KbplacerDialog(wx.Dialog):
    def __init__(
        self,
        parent: Optional[wx.Window],
        title: str,
        initial_state: WindowState = WindowState(),
    ) -> None:
        style = wx.DEFAULT_DIALOG_STYLE
        super(KbplacerDialog, self).__init__(parent, -1, title, style=style)

        language = get_current_kicad_language()
        logger.info(f"Language: {language}")
        self._ = get_plugin_translator(language)

        switch_section = self.get_switch_section(
            layout_path=initial_state.layout_path,
            key_distance=initial_state.key_distance,
            element_info=initial_state.key_info,
        )

        switch_diodes_section = self.get_switch_diodes_section(
            enable=initial_state.enable_diode_placement,
            route_switches_with_diodes=initial_state.route_switches_with_diodes,
            optimize_diodes_orientation=initial_state.optimize_diodes_orientation,
            element_info=initial_state.diode_info,
        )

        additional_elements_section = self.get_additional_elements_section(
            elements_info=initial_state.additional_elements,
        )

        misc_section = self.get_misc_section(
            route_rows_and_columns=initial_state.route_rows_and_columns,
            template_path=initial_state.template_path,
            generate_outline=initial_state.generate_outline,
            outline_delta=initial_state.outline_delta,
        )

        box = wx.BoxSizer(wx.VERTICAL)

        box.Add(switch_section, 0, wx.EXPAND | wx.ALL, 5)
        box.Add(switch_diodes_section, 0, wx.EXPAND | wx.ALL, 5)
        box.Add(additional_elements_section, 0, wx.EXPAND | wx.ALL, 5)
        box.Add(misc_section, 0, wx.EXPAND | wx.ALL, 5)

        buttons = self.CreateButtonSizer(wx.OK | wx.CANCEL | wx.HELP)

        if help_button := wx.FindWindowById(wx.ID_HELP, self):
            help_button.Bind(wx.EVT_BUTTON, self.on_help_button)

        box.Add(buttons, 0, wx.EXPAND | wx.ALL, 5)

        self.SetSizerAndFit(box)

    def get_switch_section(
        self,
        layout_path: str = "",
        key_distance: Tuple[float, float] = (19.05, 19.05),
        element_info: ElementInfo = ElementInfo(
            "SW{}", PositionOption.DEFAULT, ZERO_POSITION, ""
        ),
    ) -> wx.Sizer:
        layout_label = wx.StaticText(self, -1, self._("Keyboard layout file:"))
        layout_picker = get_file_picker(
            self,
            -1,
            wildcard="JSON files (*.json)|*.json|All files (*)|*",
            style=wx.FLP_USE_TEXTCTRL,
        )
        layout_picker.SetMinSize((400, -1))
        if layout_path:
            layout_picker.SetPath(layout_path)
            layout_picker.GetTextCtrl().SetInsertionPointEnd()

        key_distance_x = LabeledTextCtrl(
            self,
            wx_("Step X:"),
            value=str(key_distance[0]),
            width=5,
            validator=FloatValidator(),
        )
        key_distance_y = LabeledTextCtrl(
            self,
            wx_("Step Y:"),
            value=str(key_distance[1]),
            width=5,
            validator=FloatValidator(),
        )

        key_annotation = LabeledTextCtrl(
            self, wx_("Footprint Annotation") + ":", element_info.annotation_format
        )

        key_position = ElementPositionWidget(self, ZERO_POSITION, disable_offsets=True)
        if element_info.position:
            key_position.set_position(element_info.position)

        box = wx.StaticBox(self, label=self._("Switch settings"))
        sizer = wx.StaticBoxSizer(box, wx.VERTICAL)

        row1 = wx.BoxSizer(wx.HORIZONTAL)
        row1.Add(layout_label, 0, wx.LEFT | wx.RIGHT | wx.ALIGN_CENTER_VERTICAL, 5)
        row1.Add(layout_picker, 1, wx.ALL, 5)
        row1.Add(key_distance_x, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        row1.Add(key_distance_y, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer.Add(row1, 0, wx.EXPAND | wx.ALL, 5)

        row2 = wx.BoxSizer(wx.HORIZONTAL)
        row2.Add(key_annotation, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        row2.Add(key_position, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer.Add(row2, 0, wx.EXPAND | wx.ALL, 5)

        self.__layout_picker = layout_picker
        self.__key_distance_x = key_distance_x.text
        self.__key_distance_y = key_distance_y.text
        self.__key_annotation_format = key_annotation.text
        self.__key_position = key_position

        return sizer

    def get_switch_diodes_section(
        self,
        enable: bool = True,
        route_switches_with_diodes: bool = True,
        optimize_diodes_orientation: bool = False,
        element_info: ElementInfo = ElementInfo(
            "D{}", PositionOption.DEFAULT, DEFAULT_DIODE_POSITION, ""
        ),
    ) -> wx.Sizer:
        place_diodes_checkbox = wx.CheckBox(self, label=wx_("Allow autoplacement"))
        place_diodes_checkbox.SetValue(enable)
        place_diodes_checkbox.Bind(wx.EVT_CHECKBOX, self.on_diode_place_checkbox)

        switches_and_diodes_tracks_checkbox = wx.CheckBox(
            self, label=self._("Route with switches")
        )
        switches_and_diodes_tracks_checkbox.SetValue(route_switches_with_diodes)

        optimize_diodes_orientation_checkbox = wx.CheckBox(
            self, label=self._("Automatically adjust orientation")
        )
        optimize_diodes_orientation_checkbox.SetValue(optimize_diodes_orientation)

        diode_settings = ElementSettingsWidget(
            self,
            element_info,
            default_position=DEFAULT_DIODE_POSITION,
        )

        box = wx.StaticBox(self, label=self._("Switch diodes settings"))
        sizer = wx.StaticBoxSizer(box, wx.VERTICAL)

        row1 = wx.BoxSizer(wx.HORIZONTAL)
        row1.Add(place_diodes_checkbox, 0, wx.ALL, 0)
        row1.Add(switches_and_diodes_tracks_checkbox, 0, wx.ALL, 0)
        row1.Add(optimize_diodes_orientation_checkbox, 0, wx.ALL, 0)

        sizer.Add(row1, 0, wx.EXPAND | wx.ALL, 5)
        # weird border value to make it aligned with 'additional_elements_section':
        sizer.Add(diode_settings, 0, wx.EXPAND | wx.ALL, 9)

        self.__place_diodes_checkbox = place_diodes_checkbox
        self.__switches_and_diodes_tracks_checkbox = switches_and_diodes_tracks_checkbox
        self.__optimize_diodes_orientation_checkbox = (
            optimize_diodes_orientation_checkbox
        )
        self.__diode_settings = diode_settings

        self.__enable_diode_settings(enable)

        return sizer

    def __enable_diode_settings(self, enable: bool) -> None:
        if enable:
            self.__diode_settings.Enable()
            self.__optimize_diodes_orientation_checkbox.Enable()
        else:
            self.__diode_settings.Disable()
            self.__optimize_diodes_orientation_checkbox.Disable()

    def on_diode_place_checkbox(self, event: wx.CommandEvent) -> None:
        is_checked = event.GetEventObject().IsChecked()
        self.__enable_diode_settings(is_checked)

    def get_additional_elements_section(
        self,
        elements_info: List[ElementInfo] = [
            ElementInfo("ST{}", PositionOption.CUSTOM, ZERO_POSITION, "")
        ],
    ) -> wx.Sizer:
        self.__additional_elements = []

        scrolled_window = wx.ScrolledWindow(self)
        scrolled_window_sizer = wx.BoxSizer(wx.VERTICAL)

        scrolled_window.SetSizer(scrolled_window_sizer)
        virtual_width, virtual_height = scrolled_window_sizer.GetMinSize()
        scrolled_window.SetVirtualSize((virtual_width, virtual_height))
        scrolled_window.SetScrollRate(0, 10)

        box = wx.StaticBox(self, label=self._("Additional elements settings"))
        sizer = wx.StaticBoxSizer(box, wx.VERTICAL)

        sizer.Add(scrolled_window, 1, wx.EXPAND | wx.ALL, 10)

        dialog_width, _ = self.GetSize()
        sizer.SetMinSize((dialog_width, 180))

        buttons_sizer = wx.BoxSizer(wx.HORIZONTAL)

        def add_element(element_info: ElementInfo) -> None:
            element_settings = ElementSettingsWidget(scrolled_window, element_info)
            self.__additional_elements.append(element_settings)
            scrolled_window_sizer.Add(element_settings, 0, wx.EXPAND | wx.ALIGN_LEFT, 0)
            self.GetTopLevelParent().Layout()

        def add_element_callback(_) -> None:
            add_element(ElementInfo("", PositionOption.CUSTOM, ZERO_POSITION, ""))

        for element_info in elements_info:
            add_element(element_info)

        add_icon = PyEmbeddedImage(
            b"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAACtJ"
            b"REFUOI1jYKAx+A/FOAETpTaMGsDAwAil8YY0Pv0Uu4AQGE0H9DCAYgAADfAFFDV6vY8AAAAA"
            b"SUVORK5CYII="
        ).GetBitmap()
        add_button = wx.BitmapButton(self, bitmap=add_icon)
        add_button.Bind(wx.EVT_BUTTON, add_element_callback)

        def remove_element(_) -> None:
            element_settings = (
                self.__additional_elements.pop() if self.__additional_elements else None
            )
            if element_settings:
                element_settings.Destroy()
                self.Layout()

        remove_icon = PyEmbeddedImage(
            b"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABtJ"
            b"REFUOI1jYBgFwwAwQun/5OpnopZLRsGQBgBLTwEEpzJYVwAAAABJRU5ErkJggg=="
        ).GetBitmap()
        remove_button = wx.BitmapButton(self, bitmap=remove_icon)
        remove_button.Bind(wx.EVT_BUTTON, remove_element)

        buttons_sizer.Add(add_button, 0, wx.EXPAND | wx.ALL, 0)
        buttons_sizer.Add(remove_button, 0, wx.EXPAND | wx.ALL, 0)

        sizer.Add(buttons_sizer, 0, wx.EXPAND | wx.ALL, 5)

        return sizer

    def get_misc_section(
        self,
        route_rows_and_columns: bool = True,
        template_path: str = "",
        generate_outline: bool = False,
        outline_delta: float = 0.0,
    ) -> wx.Sizer:
        row_and_columns_tracks_checkbox = wx.CheckBox(
            self, label=self._("Route rows and columns")
        )
        row_and_columns_tracks_checkbox.SetValue(route_rows_and_columns)

        template_label = wx.StaticText(
            self, -1, self._("Controller circuit template file:")
        )
        template_picker = get_file_picker(
            self,
            -1,
            wildcard="KiCad printed circuit board files (*.kicad_pcb)|*.kicad_pcb",
            style=wx.FLP_USE_TEXTCTRL,
        )
        if template_path:
            template_picker.SetPath(template_path)
            template_picker.GetTextCtrl().SetInsertionPointEnd()

        row1 = wx.BoxSizer(wx.HORIZONTAL)
        row1.Add(row_and_columns_tracks_checkbox, 0, wx.EXPAND | wx.ALL, 5)
        row1.Add(wx.StaticLine(self, style=wx.LI_VERTICAL), 0, wx.EXPAND | wx.ALL, 5)
        row1.Add(template_label, 0, wx.LEFT | wx.RIGHT | wx.ALIGN_CENTER_VERTICAL, 5)
        row1.Add(template_picker, 1, wx.EXPAND | wx.ALL, 5)

        generate_outline_checkbox = wx.CheckBox(
            self, label=self._("Build board outline")
        )
        generate_outline_checkbox.SetValue(generate_outline)
        generate_outline_checkbox.Bind(
            wx.EVT_CHECKBOX, self.on_generate_outline_checkbox
        )
        outline_delta_ctrl = LabeledTextCtrl(
            self,
            self._("Outline delta:"),
            value=str(outline_delta),
            width=5,
            validator=FloatValidator(),
        )

        row2 = wx.BoxSizer(wx.HORIZONTAL)
        row2.Add(generate_outline_checkbox, 0, wx.EXPAND | wx.ALL, 5)
        row2.Add(wx.StaticLine(self, style=wx.LI_VERTICAL), 0, wx.EXPAND | wx.ALL, 5)
        row2.Add(outline_delta_ctrl, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)

        box = wx.StaticBox(self, label=self._("Other settings"))
        sizer = wx.StaticBoxSizer(box, wx.VERTICAL)
        sizer.Add(row1, 0, wx.EXPAND | wx.ALL, 5)
        sizer.Add(row2, 0, wx.EXPAND | wx.ALL, 5)

        self.__rows_and_columns_tracks_checkbox = row_and_columns_tracks_checkbox
        self.__template_picker = template_picker
        self.__generate_outline_checkbox = generate_outline_checkbox
        self.__outline_delta_ctrl = outline_delta_ctrl

        self.__enable_outline_delta(generate_outline)

        return sizer

    def __enable_outline_delta(self, enable: bool) -> None:
        if enable:
            self.__outline_delta_ctrl.Enable()
        else:
            self.__outline_delta_ctrl.Disable()

    def on_generate_outline_checkbox(self, event: wx.CommandEvent) -> None:
        is_checked = event.GetEventObject().IsChecked()
        self.__enable_outline_delta(is_checked)

    def on_help_button(self, event: wx.CommandEvent) -> None:
        del event
        help_dialog = HelpDialog(self)
        help_dialog.ShowModal()
        help_dialog.Destroy()

    def get_layout_path(self) -> str:
        return self.__layout_picker.GetPath()

    def get_key_annotation_format(self) -> str:
        return self.__key_annotation_format.GetValue()

    def route_switches_with_diodes(self) -> bool:
        return self.__switches_and_diodes_tracks_checkbox.GetValue()

    def optimize_diodes_orientation(self) -> bool:
        return self.__optimize_diodes_orientation_checkbox.GetValue()

    def route_rows_and_columns(self) -> bool:
        return self.__rows_and_columns_tracks_checkbox.GetValue()

    def get_key_distance(self) -> Tuple[float, float]:
        x = float(self.__key_distance_x.GetValue())
        y = float(self.__key_distance_y.GetValue())
        return x, y

    def get_template_path(self) -> str:
        return self.__template_picker.GetPath()

    def get_key_info(self) -> ElementInfo:
        return ElementInfo(
            self.get_key_annotation_format(),
            PositionOption.DEFAULT,
            self.__key_position.GetValue(),
            "",
        )

    def enable_diode_placement(self) -> bool:
        return self.__place_diodes_checkbox.GetValue()

    def get_diode_info(self) -> ElementInfo:
        return self.__diode_settings.GetValue()

    def get_additional_elements_info(self) -> List[ElementInfo]:
        return [e.GetValue() for e in self.__additional_elements]

    def generate_outline(self) -> bool:
        return self.__generate_outline_checkbox.GetValue()

    def get_outline_delta(self) -> float:
        return float(self.__outline_delta_ctrl.text.GetValue())

    def get_window_state(self) -> WindowState:
        return WindowState(
            layout_path=self.get_layout_path(),
            key_distance=self.get_key_distance(),
            key_info=self.get_key_info(),
            enable_diode_placement=self.enable_diode_placement(),
            route_switches_with_diodes=self.route_switches_with_diodes(),
            optimize_diodes_orientation=self.optimize_diodes_orientation(),
            diode_info=self.get_diode_info(),
            additional_elements=self.get_additional_elements_info(),
            route_rows_and_columns=self.route_rows_and_columns(),
            template_path=self.get_template_path(),
            generate_outline=self.generate_outline(),
            outline_delta=self.get_outline_delta(),
        )


def load_window_state_from_log(filepath: str) -> WindowState:
    try:
        with open(filepath, "r") as f:
            for line in f:
                if "GUI state:" in line:
                    state = WindowState.from_dict(json.loads(line[line.find("{") :]))
                    logger.info("Using window state found in previous log")
                    return state
    except Exception:
        # if something went wrong use default
        pass
    logger.warning("Failed to parse window state from log file, using default")
    return WindowState()


# used for tests
if __name__ == "__main__":
    import argparse
    import threading

    from .kbplacer_plugin import run_from_gui

    parser = argparse.ArgumentParser(description="dialog test")
    parser.add_argument(
        "-i", "--initial-state-file", default="", help="Initial gui state file"
    )
    parser.add_argument(
        "-o", "--output-dir", required=True, help="Directory for output files"
    )
    parser.add_argument(
        "--run-without-dialog",
        required=False,
        action="store_true",
        help="Run with loaded state without displaying dialog window",
    )
    parser.add_argument(
        "-b",
        "--board",
        required=False,
        default="",
        help=".kicad_pcb file to be used with --run-without-dialog option",
    )
    args = parser.parse_args()

    initial_state = load_window_state_from_log(args.initial_state_file)
    if not args.run_without_dialog:
        app = wx.App()
        dlg = KbplacerDialog(None, "kbplacer", initial_state=initial_state)

        if "PYTEST_CURRENT_TEST" in os.environ:
            print(f"Using {wx.version()}")

            # use stdin for gracefully closing GUI when running
            # from pytest. This is required when measuring
            # coverage and process kill would cause measurement to be lost
            def listen_for_exit() -> None:
                input("Press any key to exit: ")
                dlg.Close()
                wx.Exit()

            input_thread = threading.Thread(target=listen_for_exit)
            input_thread.start()

            dlg.Show()
            app.MainLoop()
        else:
            dlg.ShowModal()

        with open(f"{args.output_dir}/window_state.json", "w") as f:
            f.write(f"{dlg.get_window_state()}")

        print("exit ok")
    else:
        import pcbnew

        board = run_from_gui(args.board, initial_state)
        pcbnew.Refresh()
        pcbnew.SaveBoard(args.board, board)



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Tuple

import pcbnew

from .board_builder import BoardBuilder
from .edge_generator import build_board_outline
from .element_position import ElementInfo, PositionOption
from .kbplacer_dialog import WindowState
from .key_placer import KeyPlacer
from .template_copier import copy_from_template_to_board


@dataclass
class PluginSettings:
    board_path: str
    layout_path: str
    key_info: ElementInfo
    key_distance: Tuple[float, float]
    diode_info: ElementInfo
    route_switches_with_diodes: bool
    optimize_diodes_orientation: bool
    route_rows_and_columns: bool
    additional_elements: List[ElementInfo]
    generate_outline: bool
    outline_delta: float
    template_path: str
    create_from_annotated_layout: bool
    switch_footprint: str
    diode_footprint: str


def run(settings: PluginSettings) -> pcbnew.BOARD:
    if settings.create_from_annotated_layout:
        builder = BoardBuilder(
            settings.board_path,
            switch_footprint=settings.switch_footprint,
            diode_footprint=settings.diode_footprint,
        )
        board = builder.create_board(settings.layout_path)
    else:
        board = pcbnew.LoadBoard(settings.board_path)

    placer = KeyPlacer(board, settings.key_distance)
    placer.run(
        settings.layout_path,
        settings.key_info,
        settings.diode_info,
        settings.route_switches_with_diodes,
        settings.route_rows_and_columns,
        additional_elements=settings.additional_elements,
        optimize_diodes_orientation=settings.optimize_diodes_orientation,
    )

    if settings.generate_outline:
        build_board_outline(
            board, settings.outline_delta, settings.key_info.annotation_format
        )

    if settings.template_path:
        copy_from_template_to_board(
            board, settings.template_path, settings.route_rows_and_columns
        )

    return board


def run_from_gui(board_path: str, state: WindowState) -> pcbnew.BOARD:
    """Same as 'run' but with additional WindowState to PluginSettings translation"""
    if not state.enable_diode_placement:
        state.diode_info.position_option = PositionOption.UNCHANGED
        state.diode_info.template_path = ""

    settings = PluginSettings(
        board_path=board_path,
        layout_path=state.layout_path,
        key_info=state.key_info,
        key_distance=state.key_distance,
        diode_info=state.diode_info,
        route_switches_with_diodes=state.route_switches_with_diodes,
        optimize_diodes_orientation=state.optimize_diodes_orientation,
        route_rows_and_columns=state.route_rows_and_columns,
        additional_elements=state.additional_elements,
        generate_outline=state.generate_outline,
        outline_delta=state.outline_delta,
        template_path=state.template_path,
        create_from_annotated_layout=False,
        switch_footprint="",
        diode_footprint="",
    )
    return run(settings)
